from flask import request

def make_key():
    """A function which is called to derive the key for a computed value.
       The key in this case is the concat value of all the json request
       parameters. Other strategy could to use any hashing function.
    :returns: unique string for which the value should be cached.
    """
    user_data = request.get_json()
    # app.logger.info(f'make_key(): {user_data}')
    return ",".join([f"{key}={value}" for key, value in user_data.items()])



import overpass
import pandas as pd
import geopandas as gpd
import mapbox_earcut as earcut
import numpy as np
import re
import time
import vedo
import osmium as o
import subprocess
import pyproj
import matplotlib.pyplot as plt

from geopy.geocoders import Nominatim
from shapely.geometry import MultiPolygon, Polygon, MultiLineString, LineString, box, Point
from shapely.ops import linemerge, transform
from shapely.wkb import loads
from shapely.validation import explain_validity

from . import utils
from . import errors
from . import cache
from .buildings import Buildings
from .urban_component import UrbanComponent


class RelationHandler(o.SimpleHandler):
    def __init__(self, filters):
        o.SimpleHandler.__init__(self)
        self.relation_elements = {'elements':[]}
        self.relations_position = {}
        self.relation_ways_ids = [] # all ids of ways that are part of relations
        self.filters = filters

    def node(self, n):
        pass

    def way(self, w):
        pass

    def relation(self, r):

        tags = {}
        pass_filters = False
        disqualified = False

        for tag in r.tags:
            tags[tag.k] = tag.v

            if('rel' in self.filters and tag.k in self.filters['rel'] and (tag.v in self.filters['rel'][tag.k] or -1 in self.filters['rel'][tag.k])): # -1 includes all
                pass_filters = True

            if('rel' in self.filters and 'disqualifiers' in self.filters['rel'] and tag.k in self.filters['rel']['disqualifiers'] and (tag.v in self.filters['rel']['disqualifiers'][tag.k] or -1 in self.filters['rel']['disqualifiers'][tag.k])): # -1 includes all
                disqualified = True

        if(pass_filters and not disqualified):

            members = []

            for member in r.members:

                type = member.type

                if type == 'w':
                    type = 'way'

                self.relation_ways_ids.append(member.ref)

                members.append({
                    'id': member.ref,
                    'type': type,
                    'role': member.role,
                    'geometry': []
                })

            self.relation_elements['elements'].append({
                'type': 'relation',
                'id': r.id,
                'members': members,
                'bounds': None,
                'tags': tags
            })

            self.relations_position[r.id] = len(self.relation_elements['elements'])-1

    def area(self, a):
        pass

class OSMHandler(o.SimpleHandler):
    def __init__(self, filters, relation_ways_ids):
        o.SimpleHandler.__init__(self)
        self.relation_ways_ids = relation_ways_ids

        self.ways_elements = {'elements':[]}
        self.ways_elements_of_relations = {'elements':[]}
        self.areas = {}
        self.ways_position = {}
        self.filters = filters
    
    def node(self, n):
        pass

    def way(self, w):

        tags = {}
        pass_filters = False
        disqualified = False

        for tag in w.tags:
            tags[tag.k] = tag.v

            if('way' in self.filters and tag.k in self.filters['way'] and (tag.v in self.filters['way'][tag.k] or -1 in self.filters['way'][tag.k])):
                pass_filters = True

            if('way' in self.filters and 'disqualifiers' in self.filters['way'] and tag.k in self.filters['way']['disqualifiers'] and (tag.v in self.filters['way']['disqualifiers'][tag.k] or -1 in self.filters['way']['disqualifiers'][tag.k])): # -1 includes all
                disqualified = True

        if((pass_filters and not disqualified) or w.id in self.relation_ways_ids):

            nodes_ids = []
            geometry = []
            bounds = {
                'minlat': None,
                'minlon': None,
                'maxlat': None, 
                'maxlon': None
            }

            for elem in w.nodes:
                nodes_ids.append(elem.ref)

                geometry.append({
                    'lat': elem.lat,
                    'lon': elem.lon
                })

                if(bounds['minlat'] == None):
                    bounds['minlat'] = elem.lat
                elif(elem.lat < bounds['minlat']):
                        bounds['minlat'] = elem.lat

                if(bounds['minlon'] == None):
                    bounds['minlon'] = elem.lon
                elif(elem.lon < bounds['minlon']):
                        bounds['minlon'] = elem.lon

                if(bounds['maxlat'] == None):
                    bounds['maxlat'] = elem.lat
                elif(elem.lat > bounds['maxlat']):
                        bounds['maxlat'] = elem.lat

                if(bounds['maxlon'] == None):
                    bounds['maxlon'] = elem.lon
                elif(elem.lon > bounds['maxlon']):
                        bounds['maxlon'] = elem.lon

            if(w.id in self.relation_ways_ids):

                self.ways_elements_of_relations['elements'].append({
                    'type': 'way',
                    'id': w.id,
                    'bounds': bounds,
                    'nodes': nodes_ids,
                    'geometry': geometry,
                    'tags': tags
                })
            
                self.ways_position[w.id] = len(self.ways_elements_of_relations['elements'])-1
            
            else:
                self.ways_elements['elements'].append({
                    'type': 'way',
                    'id': w.id,
                    'bounds': bounds,
                    'nodes': nodes_ids,
                    'geometry': geometry,
                    'tags': tags
                })


    def relation(self, r):
        pass

    def area(self, a):
        pass

class OSM:

    def load(region, layers=['parks','water','roads','buildings'], pbf_filepath=None):
        '''
            Region can be a bounding polygon, a bounding box or an address
        '''

        if(isinstance(region, str)): # address
            return OSM.load_from_address(region, layers, pbf_filepath)
        elif(len(region) == 4 and (isinstance(region[0], float) or isinstance(region[0], int))): # bounding box
            return OSM.load_from_bbox(region, layers, pbf_filepath)
        elif(len(region[0]) == 2): # polygon
            return OSM.load_from_polygon(region, layers)
        else:
            raise Exception("Region format "+str(region)+" not supported")

    def load_from_bbox(bbox, layers=['parks','water','roads','buildings'], pbf_filepath=None):
        '''
            Load layers inside bounding box to memory storing them into the UrbanComponent

            Args:
                bbox (float[]): List containing the two extreme corners of the bounding box. [minLat, minLng, maxLat, maxLng]
                layers (string[]): Name of the layers that will be loaded. Possible values: parks, water, costline, roads, buildings. If buildings are used it must be the last layer to ensure correct rendering.
                    (default is ['parks', 'water', 'roads','buildings'])
                filepath (string): Location of the pbf file to load. This argument is optional. If provided the data will be loaded from the pbf instead of the OSM API
                    (default is None)

            Returns:
                component (UrbanComponent): Allows the manipulation of the loaded data
        
        '''
        
        cam = utils.get_camera(bbox, True)
        
        loaded = None

        if(pbf_filepath != None):

            aux = '%f,%f,%f,%f'%(bbox[1],bbox[0],bbox[3],bbox[2])

            split_file = pbf_filepath.split('/')

            output_file = ''

            for index, section in enumerate(split_file):
                if(index == len(split_file)-1):
                    output_file += 'filtered_'+section
                else:
                    output_file += section+'/'

            proc = subprocess.call(['wsl', 'osmium', 'extract', '-b', aux, '-o', output_file, '--overwrite', pbf_filepath], shell=True) # TODO: make it not depend on the OS

            loaded = OSM.get_osm(bbox, True, layers, output_file)
        else:
            loaded = OSM.get_osm(bbox, True, layers)

        component = UrbanComponent(layers = loaded, bpolygon = bbox, camera = cam)

        return component

    def load_from_polygon(bpolygon, layers=['parks','water','roads','buildings']):
        
        flattened_polygon = [item for row in bpolygon for item in row]

        # cam = utils.get_camera(bpolygon)
        cam = utils.get_camera(flattened_polygon)

        # loaded = OSM.get_osm(bpolygon, False, layers)
        loaded = OSM.get_osm(flattened_polygon, False, layers)

        # component = UrbanComponent(layers = loaded, bpolygon = bpolygon, camera = cam)
        component = UrbanComponent(layers = loaded, bpolygon = flattened_polygon, camera = cam)

        return component

    def load_from_address(address, layers=['parks','water','roads','buildings'], pbf_filepath=None):

        geolocator = Nominatim(user_agent="urbantk")

        location = geolocator.geocode(address, timeout=5).raw

        bbox = [float(x) for x in location['boundingbox']]
        bbox = [bbox[0],bbox[2],bbox[1],bbox[3]]

        if(pbf_filepath != None):
            return OSM.load_from_bbox(bbox, layers, pbf_filepath)
        else:
            return OSM.load_from_bbox(bbox, layers)

    def get_osm(bpolygon, bbox=False, layers=['parks','water','roads','buildings'], pbf_filepath=None):

        '''
            Request data to OSM API using overpass and builds meshes for each loaded data from the result

            Args:
                bpolygon (float[]): coordinates of bounding polygon
                bbox: if bpoly follows the format [minLat, minLng, maxLat, maxLng]
                layers (string[]): Name of the layers that will be loaded. Possible values: parks, water, costline, roads, buildings. If buildings are used it must be the last layer to ensure correct rendering.
                    (default is ['parks', 'water', 'roads','buildings'])
                filepath (string): Location of the pbf file to load. This argument is optional. If provided the data will be loaded from the pbf instead of the OSM API
                    (default is None)

            Returns:
                result (list[object]): A list of python objects representing the layers in json format
        '''

        bpoly = bpolygon.copy()
        api = overpass.API()

        overpass_responses = {}

        for layer_obj in layers:

            layer = ''

            if isinstance(layer_obj, str):
                layer = layer_obj
            else:
                layer = layer_obj['name']

            if layer == 'surface':
                continue
            
            if(pbf_filepath == None):

                query = OSM.build_osm_query(bpoly, 'geom', bbox, [layer])

                response = cache._load_osm_from_cache(query)

                if not response:
                    time.sleep(1) # avoiding Overpass 429 Too Many Requests
                    response = api.get(query, build=False)
                    cache._save_osm_to_cache(query,response)

                overpass_responses[layer] = OSM.parse_osm(response)
            else:
                relation_handler = RelationHandler(OSM.get_osmium_filters(layer))

                relation_handler.apply_file(pbf_filepath, locations=True)

                relation_elements = relation_handler.relation_elements
                relations_position = relation_handler.relations_position
                relation_ways_ids = relation_handler.relation_ways_ids

                osmhandler = OSMHandler(OSM.get_osmium_filters(layer), relation_ways_ids)

                osmhandler.apply_file(pbf_filepath, locations=True)

                ways_position = osmhandler.ways_position
                ways_elements = osmhandler.ways_elements
                ways_elements_of_relations = osmhandler.ways_elements_of_relations
                areas = osmhandler.areas

                complete_relation_elements = OSM.fill_relation_geom_osmium(ways_elements_of_relations, relation_elements, ways_position, relations_position, areas)

                overpass_responses[layer] = OSM.parse_osm({'elements': ways_elements['elements'] + complete_relation_elements['elements']})

        result = []
        result_gdf_objects = []
        result_gdf_coordinates = []
        result_gdf_coordinates_3d = []
        result_gdf_perSection = [] # for 3D buildings. Each row represents a section of the building
        ttype = ''
        styleKey = ''
        renderStyle = []

        for layer_obj in layers:

            layer = ''
            args = {}

            if isinstance(layer_obj, str):
                layer = layer_obj
            else:
                layer = layer_obj['name']
                args = layer_obj['args']

            if layer == 'surface':
                nCells = -1
                sizeCells = -1

                if 'nCells' in args:
                    nCells = args['nCells']
                
                if 'sizeCells' in args:
                    sizeCells = args['sizeCells']

                layer_geometry = OSM.create_surface_mesh(bpoly, bbox, nCells, sizeCells)
                geometry = layer_geometry['data']
                result_gdf_objects.append(layer_geometry['gdf']['objects'])
                result_gdf_coordinates.append(layer_geometry['gdf']['coordinates'])
                result_gdf_coordinates_3d.append(layer_geometry['gdf']['coordinates'])
                ttype = 'HEATMAP_LAYER'
                styleKey = 'surface'
                renderStyle = ['SMOOTH_COLOR_MAP']
            elif layer == 'buildings':

                sizeCells = -1 # default cell size for buildings no subdivision

                if 'sizeCells' in args:
                    sizeCells = args['sizeCells']

                layer_geometry = OSM.osm_to_building_mesh(overpass_responses[layer], bpoly, bbox, sizeCells)
                geometry = layer_geometry['data']
                result_gdf_objects.append(layer_geometry['gdf']['objects'])
                result_gdf_coordinates.append(layer_geometry['gdf']['coordinates'])
                result_gdf_coordinates_3d.append(layer_geometry['gdf']['coordinates3d'])
                result_gdf_perSection.append(layer_geometry['perSectionGdf'])
                ttype = 'BUILDINGS_LAYER'
                styleKey = 'building'
                renderStyle = ['SMOOTH_COLOR_MAP_TEX']
            elif layer == 'roads':
                layer_geometry = OSM.osm_to_roads_polyline(overpass_responses[layer], bpoly, bbox)
                geometry = layer_geometry['data']
                result_gdf_objects.append(layer_geometry['gdf']['objects'])
                result_gdf_coordinates.append(layer_geometry['gdf']['coordinates'])
                result_gdf_coordinates_3d.append(layer_geometry['gdf']['coordinates3d'])
                ttype = 'TRIANGLES_3D_LAYER'
                styleKey = 'roads'
                renderStyle = ['FLAT_COLOR']
            elif layer == 'coastline':
                layer_geometry = OSM.osm_to_coastline_mesh(overpass_responses[layer], bpoly, bbox)
                geometry = layer_geometry['data']
                result_gdf_objects.append(layer_geometry['gdf']['objects'])
                result_gdf_coordinates.append(layer_geometry['gdf']['coordinates'])
                result_gdf_coordinates_3d.append(layer_geometry['gdf']['coordinates3d'])
                ttype = 'TRIANGLES_3D_LAYER'
                styleKey = 'land'
                renderStyle = ['FLAT_COLOR']
            else:
                layer_geometry = OSM.osm_to_generic_mesh(overpass_responses[layer], bpoly, bbox, convert2dto3d=True)
                geometry = layer_geometry['data']
                result_gdf_objects.append(layer_geometry['gdf']['objects'])
                result_gdf_coordinates.append(layer_geometry['gdf']['coordinates'])
                result_gdf_coordinates_3d.append(layer_geometry['gdf']['coordinates3d'])
                ttype = 'TRIANGLES_3D_LAYER'
                styleKey = layer
                renderStyle = ['FLAT_COLOR']

            result.append({'id': layer, 'type': ttype, 'renderStyle': renderStyle, 'styleKey': styleKey, 'data': geometry})
        
        return {'json': result, 'gdf': {'sections': result_gdf_perSection, 'objects': result_gdf_objects, 'coordinates': result_gdf_coordinates, 'coordinates3d': result_gdf_coordinates_3d}}

    def osm_to_roads_polyline(osm_elements, bpoly, bbox):
        '''
            Creates the roads polyline based on the OSM elements

            Args:
                osm_elements (object): A json object describing the components of the roads layer 
                bpoly (float[]): coordinates of bounding polygon
                bbox: if bpoly follows the format [minLat, minLng, maxLat, maxLng]

            Returns:
                mesh (object): A json object describing the geometry of the layer
        '''

        mesh = []
        coords = []
        for wid in osm_elements['ways']:
            way = osm_elements['ways'][wid]
            coords.append(way['geometry'])
        
        lines = MultiLineString(coords)
        inter = lines.intersection(utils.polygon_bpoly(bpoly, bbox))

        proj_4326 = pyproj.CRS('EPSG:4326')
        proj_3395 = pyproj.CRS('EPSG:3395')
        project = pyproj.Transformer.from_crs(proj_4326, proj_3395, always_xy=True).transform

        geometries = []
        geometries_coordinates = []

        ids = []
        ids_coordinates = []
        counter_id_coordinates = 0

        for id, line in enumerate(inter):

            # extruding line ============================
            x, y = line.coords.xy
            invertedLine = LineString(list(zip(y,x)))
            
            ids.append(id)

            transformed_line = transform(project, invertedLine)

            buffer_line = transformed_line.buffer(2) # in meters
            # buffer_line = transformed_line.buffer(7.5) # in meters

            geometries.append(buffer_line)

            x, y = buffer_line.exterior.coords.xy

            nodes = list(zip(x,y))
            rings = [len(nodes)]

            indices = earcut.triangulate_float64(nodes, rings)

            nodes = np.array(nodes)
            
            # empty triangulation
            if(len(indices) == 0 or (len(indices) % 3) > 0):
                raise errors.InvalidPolygon('Invalid triangulation')

            nodes = nodes.flatten().tolist()
            indices = indices.tolist()

            for i in range(int(len(nodes)/2)):
                geometries_coordinates.append(Point(nodes[i*2], nodes[i*2+1]))
                ids_coordinates.append(counter_id_coordinates)
                counter_id_coordinates += 1

            nodes = utils.from_2d_to_3d(nodes)

            mesh.append({'type': 'type', 'geometry': {'coordinates': [round(item,4) for item in nodes], 'indices': indices}})

        gdf = gpd.GeoDataFrame({'geometry': geometries, 'id': ids}, crs=3395)

        gdf_coordinates = gpd.GeoDataFrame({'geometry': geometries_coordinates, 'id': ids_coordinates}, crs=3395)

        return {'data': mesh, 'gdf': {'objects': gdf, 'coordinates': gdf_coordinates, 'coordinates3d': None}}

    def osm_to_coastline_mesh(osm_elements, bpoly, using_bbox):

        '''
            Creates the coastline mesh based on the OSM elements

            Args:
                osm_elements (object): A json object describing the components of the coastline layer 
                bpoly (float[]): coordinates of bounding polygon
                bbox: if bpoly follows the format [minLat, minLng, maxLat, maxLng]

            Returns:
                mesh (object): A json object describing the geometry of the layer
        '''

        bbox = []

        if(using_bbox):
            bbox = bpoly
        else:
            bbox = utils.polygon_bpoly(bpoly).bounds

        # to shapely
        lines = []
        for wid in osm_elements['ways']:
            way = osm_elements['ways'][wid]
            linestring = LineString(way['geometry'])
            lines.append(linestring)
        multiline = MultiLineString(lines)

        merged_lines = linemerge(multiline)

        sides = { \
            'left': LineString([(bbox[0],bbox[1]),(bbox[2],bbox[1])]), \
            'right': LineString([(bbox[0],bbox[3]),(bbox[2],bbox[3])]), \
            'top': LineString([(bbox[2],bbox[1]),(bbox[2],bbox[3])]),\
            'bottom': LineString([(bbox[0],bbox[1]),(bbox[0],bbox[3])]) \
        }

        def test_side(line):
            for side in sides:
                if line.intersects(sides[side]):
                    return (side,list(line.intersection(sides[side]).coords)[0])


        # look for exits and enters into bounding box
        enters_byside = {'left': [], 'right': [], 'top': [], 'bottom': []}
        exits_byside = {'left': [], 'right': [], 'top': [], 'bottom': []}
        enters_byline = {}
        exits_byline = {}
        lines = []
        innerlines = []

        if merged_lines.geom_type == 'LineString':
            merged_lines = [merged_lines]

        for line in merged_lines:
            way = {'geometry': list(line.coords), 'bbox': line.bounds}
            # ignore holes for now
            if way['geometry'][0] == way['geometry'][-1]: # if the way is closed
                innerlines.append(way['geometry'])
            elif utils.intersect_bbox(bbox,way['bbox']) or utils.intersect_bbox(way['bbox'],bbox):
                curline = []
                enter_side = -1
                exit_side = -1
                for i in range(1,len(way['geometry'])):
                    p0 = way['geometry'][i-1]
                    p1 = way['geometry'][i]
                    p0_inside = utils.point_within_bbox(p0,bbox)
                    p1_inside = utils.point_within_bbox(p1,bbox)
                    # entering
                    if not p0_inside and p1_inside:
                        intersection = test_side(LineString([p0,p1]))
                        enter_side = intersection[0]
                        curline.append(intersection[1])
                        curline.append(p1)
                    # exiting
                    elif p0_inside and not p1_inside:
                        intersection = test_side(LineString([p0,p1]))
                        exit_side = intersection[0]
                        curline.append(intersection[1])

                        # found complete line, add to lines
                        lines.append(curline)
                        enters_byside[enter_side].append((curline[0],len(lines)-1))
                        enters_byline[len(lines)-1] = (curline[0],enter_side)
                        exits_byside[exit_side].append((curline[-1],len(lines)-1))
                        exits_byline[len(lines)-1] = (curline[-1],exit_side)
                        curline = []
                    # inside bbox
                    elif p0_inside and p1_inside:
                        curline.append(p0)
        
        # sort enter and exits from left to right, bottom to top q
        enters_byside['left'].sort(key = lambda x: x[1])
        exits_byside['left'].sort(key = lambda x: x[1])
        enters_byside['right'].sort(reverse=True,key = lambda x: x[1])
        exits_byside['right'].sort(reverse=True,key = lambda x: x[1])
        enters_byside['top'].sort(key = lambda x: x[0])
        exits_byside['top'].sort(key = lambda x: x[0])
        enters_byside['bottom'].sort(reverse=True,key = lambda x: x[0])
        exits_byside['bottom'].sort(reverse=True,key = lambda x: x[0])

        def next_enter(side,exit_coord):
            next_enter = -1
            for i in range(0,len(enters_byside[side])):
                coord = enters_byside[side][i][0]
                if side =='left' and coord[0] > exit_coord[0]:
                    next_enter = enters_byside[side][i][1]
                    break
                elif side == 'top' and coord[1] > exit_coord[1]:
                    next_enter = enters_byside[side][i][1]
                    break
                elif side == 'right' and coord[0] < exit_coord[0]:
                    next_enter = enters_byside[side][i][1]
                    break
                elif side == 'bottom' and coord[1] < exit_coord[1]:
                    next_enter = enters_byside[side][i][1]
                    break
            return next_enter

        def next_corner(side):
            if side == 'left':
                return (bbox[2],bbox[1])
            elif side == 'right':
                return (bbox[0],bbox[3])
            elif side == 'top':
                return (bbox[2],bbox[3])
            elif side == 'bottom':
                return (bbox[0],bbox[1])

        def next_side(side):
            if side == 'left':
                return 'top'
            elif side == 'right':
                return 'bottom'
            elif side == 'top':
                return 'right'
            elif side == 'bottom':
                return 'left'

        def next_line(curlineid,firstlineid,line):
            enter_side = enters_byline[curlineid][1]
            enter_coord = enters_byline[curlineid][0]
            exit_side = exits_byline[curlineid][1]
            exit_coord = exits_byline[curlineid][0]

            if len(line) > 0 and curlineid == firstlineid:
                return
            
            line.extend(lines[curlineid])
            # get next closest enter to the right of exit, else get corner
            if len(enters_byside[exit_side]) > 0:
                nextt = next_enter(exit_side,exit_coord)
                if(nextt != -1): # if there is a next line
                    next_line(nextt,firstlineid,line)
            else:
                nextcoord = next_corner(exit_side)
                nextside = next_side(exit_side)
                nextt = next_enter(nextside,nextcoord)
                if(nextt != -1): # if there is a next line
                    line.append(nextcoord) # dont know if this is on the right place
                    next_line(nextt,firstlineid,line)

        # stitch
        polygons = []
        for lineid in enters_byline:
            line = []
            next_line(lineid,lineid,line)
            polygons.append({'outer': line, 'inner': innerlines, 'type': 'type'})
            break

        proj_4326 = pyproj.CRS('EPSG:4326')
        proj_3395 = pyproj.CRS('EPSG:3395')
        project = pyproj.Transformer.from_crs(proj_4326, proj_3395, always_xy=True).transform

        geometries = []
        geometries_coordinates = []

        ids = []
        ids_coordinates = []
        counter_id_coordinates = 0

        # triangulate
        mesh = []
        for id, poly in enumerate(polygons):

            ids.append(id)

            nodes = []
            rings = []

            # outer
            nodes = poly['outer']
            rings.append(len(nodes))

            inner_lng_lat = []

            # inner
            for inner in poly['inner']:
                inner_lng_lat.append([(elem[1], elem[0]) for elem in inner])
                nodes.extend(inner)
                rings.append(len(nodes))
            
            outer_lng_lat = [(elem[1], elem[0]) for elem in poly['outer']]

            geometries.append(transform(project,Polygon(outer_lng_lat, inner_lng_lat)))

            nodes = np.array(nodes)

            indices = earcut.triangulate_float64(nodes, rings)
            indices = np.flip(indices, axis=0)

            # empty triangulation
            if(len(indices) == 0 or (len(indices) % 3) > 0):
                raise errors.InvalidPolygon('Invalid triangulation')

            # bad triangulation
            nodes = nodes.flatten().tolist()
            indices = indices.tolist()
            dev = utils.deviation(nodes, rings, 2, indices)
            # if(abs(dev) > 0.001): # TODO: prevent bad triangulation of complex meshes
            #     raise errors.InvalidPolygon('Invalid deviation (%f)'%dev)

            nodes = utils.convert_projections("4326", "3395", nodes)
            
            for i in range(int(len(nodes)/2)):
                geometries_coordinates.append(Point(nodes[i*2], nodes[i*2+1]))
                ids_coordinates.append(counter_id_coordinates)
                counter_id_coordinates += 1

            nodes = utils.from_2d_to_3d(nodes)

            mesh.append({'type': poly['type'], 'geometry': {'coordinates': [round(item,4) for item in nodes], 'indices': indices}})

        gdf = gpd.GeoDataFrame({'geometry': geometries, 'id': ids}, crs=3395)

        gdf_coordinates = gpd.GeoDataFrame({'geometry': geometries_coordinates, 'id': ids_coordinates}, crs=3395)

        return {'data': mesh, 'gdf': {'objects': gdf, 'coordinates': gdf_coordinates, 'coordinates3d': None}}

    def osm_to_building_mesh(osm_elements, bpoly, bbox = False, sizeCells = -1):
        '''
            Creates the building mesh based on the OSM elements

            Args:
                osm_elements (object): A json object describing the components of the layer 
                bpoly (float[]): coordinates of bounding polygon
                bbox: if bpoly follows the format [minLat, minLng, maxLat, maxLng]

            Returns:
                mesh (object): A json object describing the geometry of the layer
        '''

        # start = time.time()

        ways = []
        # handle multiways first
        for mid in osm_elements['multiways']:
            multiways = osm_elements['multiways'][mid]
            # https://wiki.openstreetmap.org/wiki/Relation:multipolygon
            for way in multiways:
                outernodes = []
                prevnodes = []
                inserted = False
                for outer in way['outer']:
                    curnodes = outer['geometry']
                    if curnodes[0][0]==curnodes[-1][0] and curnodes[0][1]==curnodes[-1][1]: # closed
                        ways.append({'inner': [], 'outer': curnodes, 'type': 'type'})
                        inserted = True
                    else:
                        if (len(prevnodes) == 0) or (prevnodes[-1][0]==curnodes[0][0] and prevnodes[-1][1]==curnodes[0][1]):
                            outernodes.extend(curnodes)
                        else:
                            outernodes.extend(curnodes[::-1])
                        prevnodes = curnodes
                if inserted == False:
                    ways.append({'inner': [], 'outer': outernodes, 'tags': way['tags'], 'type': 'type'})
                
                innernodes = []
                prevnodes = []
                inserted = False
                lastouter = len(ways)-1
                for inner in way['inner']:
                    curnodes = inner['geometry']
                    if curnodes[0][0]==curnodes[-1][0] and curnodes[0][1]==curnodes[-1][1]: # closed
                        ways[lastouter]['inner'].append(curnodes)
                        inserted = True
                    else:
                        if (len(prevnodes) == 0) or (prevnodes[-1][0]==curnodes[0][0] and prevnodes[-1][1]==curnodes[0][1]):
                            innernodes.append(curnodes)
                        else:
                            innernodes.extend(curnodes[::-1])
                        prevnodes = curnodes
                if inserted == False:
                    ways[lastouter]['inner'] = innernodes

        # single ways
        for wid in osm_elements['ways']:
            way = osm_elements['ways'][wid]
            nodes = way['geometry']
            ways.append({'outer': nodes, 'inner': [], 'tags': way['tags'],'type': 'type'})

        # to shapely
        polygons = []
        for way in ways:
            if len(way['outer']) > 2:
                poly = Polygon(way['outer'],way['inner'])
                poly = poly.buffer(0)
            else:
                print('small poly')
                continue
            if not poly.is_valid:
                print(explain_validity(poly))
                print('poly not valid')
                return
            if poly.overlaps(utils.polygon_bpoly(bpoly, bbox)):
                continue
            if poly.geom_type == 'Polygon':
                exterior = list(poly.exterior.coords)
                interiors = []
                for interior in poly.interiors:
                    interiors.append(list(interior.coords))
                polygons.append({'geom': [exterior, interiors], 'tags': way['tags']})
            elif poly.geom_type == 'MultiPolygon':
                for p in poly:
                    exterior = list(p.exterior.coords)
                    interiors = []
                    for interior in p.interiors:
                        interiors.append(list(interior.coords))
                    polygons.append({'geom': [exterior, interiors], 'tags': way['tags']})

        # https://wiki.openstreetmap.org/wiki/Simple_3D_buildings#Other_roof_tags
        def _feet_to_meters(s):
            # r = re.compile('^(?!$|.*\'[^\x22]+$)(?:([0-9]+)\')?(?:([0-9]+)\x22?)?$')
            r = re.compile("([0-9]*\.?[0-9]+)'([0-9]*\.?[0-9]+)?\"?") # Modified
            m = r.findall(s)[0]
            if len(m[0]) > 0 and len(m[1]) > 0:
                m = float(m[0]) + float(m[1]) / 12.0
            elif len(m[0]) > 0:
                m = float(m[0])
            return m * 0.3048
        LEVEL_HEIGHT = 3.4
        def _get_height(tags):
            if 'height' in tags:
                # already accounts for roof
                if '\'' in tags['height'] or '\"' in tags['height']:
                    return _feet_to_meters(tags['height'])
                r = re.compile(r"[-+]?\d*\.\d+|\d+")
                return float(r.findall(tags['height'])[0])
            if 'levels' in tags or 'building:levels' in tags:
                roof_height = 0
                if 'roof_height' in tags:
                    if '\'' in tags['roof_height'] or '\"' in tags['roof_height']:
                        roof_height = _feet_to_meters(tags['roof_height'])
                    else:
                        r = re.compile(r"[-+]?\d*\.\d+|\d+")
                        roof_height = float(r.findall(tags['roof_height'])[0])

                level_tag = ''

                # does not account for roof height
                if 'levels' in tags:
                    level_tag = 'levels'
                elif 'building:levels' in tags:
                    level_tag = 'building:levels'

                height = 0

                try:
                    height = float(tags[level_tag]) * LEVEL_HEIGHT
                except:
                    pass

                if 'roof_levels' in tags and roof_height == 0:
                    height += float(tags['roof_levels']) * LEVEL_HEIGHT
                return height
            return 7.0 # Modified

        def _get_min_height(tags):
            # Modified
            min_height = 0
            if 'min_height' in tags:
                # already accounts for roof
                if '\'' in tags['min_height'] or '\"' in tags['min_height']:
                    return _feet_to_meters(tags['min_height'])
                r = re.compile(r"[-+]?\d*\.\d+|\d+")
                min_height = float(r.findall(tags['min_height'])[0])
            if 'min_level' in tags:
                height = float(tags['min_level']) * LEVEL_HEIGHT
                min_height = height
            return max(0.0, min_height)

        # flip x and y coordinates
        def _invert(elem):
            return [(coord[1], coord[0]) for coord in elem]

        tags = []
        geometry = []
        min_heights = []
        heights = []
        building_id = []

        for index_polygon, building_info in enumerate(polygons):

            tags.append(building_info['tags'])
            min_heights.append(_get_min_height(building_info['tags']))
            heights.append(_get_height(building_info['tags']))

            shapely_polygons = [Polygon(_invert(elem)) for elem in building_info['geom']]
            geometry.append(MultiPolygon(shapely_polygons))

            building_id.append(index_polygon)

        geometry = gpd.GeoSeries(geometry, crs='epsg:4326')
        heights = pd.Series(heights, dtype='float')
        min_heights = pd.Series(min_heights, dtype='float')
        tags = pd.Series(tags)
        building_id = pd.Series(building_id, dtype='UInt64')

        gdf = gpd.GeoDataFrame({
            'building_id': building_id,
            'geometry': geometry,
            'min_height': min_heights,
            'height': heights,
            'tags': tags
        }, index=geometry.index)

        # drop bad values
        gdf = gdf[~gdf['height'].isnull()]
        gdf = gdf[gdf['min_height'] >= 0]
        gdf = gdf[gdf['height'] > 0]

        # new index
        gdf = gdf.set_index('building_id', drop=False)
        gdf = gdf.sort_index()

        gdf_merged_buildings = Buildings.merge_overlapping_buildings(gdf)

        return OSM.mesh_from_buildings_gdf(gdf_merged_buildings, sizeCells)

    def mesh_from_buildings_gdf(gdf, sizeCells):

        gdf = gdf.set_index('building_id', drop=False)
        gdf = gdf.sort_index()

        layer_dataframes = Buildings.generate_building_layer(gdf, sizeCells) #gdf, size

        df_mesh = layer_dataframes['df']

        json_mesh = Buildings.df_to_json(df_mesh) # prepares the layer  

        return {"data": json_mesh['data'], "perSectionGdf": gdf, "gdf": {'objects': layer_dataframes['gdf']['objects'], 'coordinates': layer_dataframes['gdf']['coordinates'], "coordinates3d": layer_dataframes['gdf']['coordinates3d']}}


    def osm_to_generic_mesh(osm_elements, bpoly, bbox, convert2dto3d=False):
        '''
            Used to load all generic layers that do not have specific functions to handle

            Args:
                osm_elements (object): A json object describing the components of layer 
                bpoly (float[]): coordinates of bounding polygon
                bbox: if bpoly follows the format [minLat, minLng, maxLat, maxLng]
                convert2dto3d (boolean): Indicates if the layer should be converted to 3D by adding z=0
                    (default is False)

            Returns:
                mesh (object): A json object describing the geometry of the layer
        '''

        ways = []
        # handle multiways first
        for mid in osm_elements['multiways']:

            multiways = osm_elements['multiways'][mid]

            # https://wiki.openstreetmap.org/wiki/Relation:multipolygon
            for way in multiways:
                outernodes = []
                prevnodes = []
                inserted = False
                for outer in way['outer']:
                    curnodes = outer['geometry']
                    if curnodes[0][0]==curnodes[-1][0] and curnodes[0][1]==curnodes[-1][1]: # closed
                        ways.append({'inner': [], 'outer': curnodes, 'type': 'type'})
                        inserted = True
                    else:
                        if (len(prevnodes) == 0) or (prevnodes[-1][0]==curnodes[0][0] and prevnodes[-1][1]==curnodes[0][1]):
                            outernodes.extend(curnodes)
                        else:
                            outernodes.extend(curnodes[::-1])
                        prevnodes = curnodes

                if inserted == False:
                    ways.append({'inner': [], 'outer': outernodes, 'type': 'type'})
                
                innernodes = []
                prevnodes = []
                inserted = False
                lastouter = len(ways)-1
                for inner in way['inner']:
                    curnodes = inner['geometry']
                    if curnodes[0][0]==curnodes[-1][0] and curnodes[0][1]==curnodes[-1][1]: # closed
                        ways[lastouter]['inner'].append(curnodes)
                        inserted = True
                    else:
                        if (len(prevnodes) == 0) or (prevnodes[-1][0]==curnodes[0][0] and prevnodes[-1][1]==curnodes[0][1]):
                            innernodes.append(curnodes)
                        else:
                            innernodes.extend(curnodes[::-1])
                        prevnodes = curnodes
                if inserted == False:
                    ways[lastouter]['inner'] = innernodes

        # single ways
        for wid in osm_elements['ways']:
            way = osm_elements['ways'][wid]
            nodes = way['geometry']
            ways.append({'outer': nodes, 'inner': [], 'type': 'type'})

        # to shapely
        polygons = []
        for way in ways:
            if len(way['outer']) > 2:
                poly = Polygon(way['outer'],way['inner'])
                poly = poly.buffer(0)
            else:
                print('small poly')
                continue
            if not poly.is_valid:
                print(explain_validity(poly))
                print('poly not valid')
                return

            poly = poly.intersection(utils.polygon_bpoly(bpoly, bbox))
            if poly.geom_type == 'Polygon':
                exterior = list(poly.exterior.coords)
                interiors = []
                for interior in poly.interiors:
                    interiors.append(list(interior.coords))
                if(len(exterior) > 0 or len(interiors) > 0):
                    polygons.append([exterior, interiors])
            elif poly.geom_type == 'MultiPolygon':
                for p in poly:
                    exterior = list(p.exterior.coords)
                    interiors = []
                    for interior in p.interiors:
                        interiors.append(list(interior.coords))
                    if(len(exterior) > 0 or len(interiors) > 0):
                        polygons.append([exterior, interiors])

        proj_4326 = pyproj.CRS('EPSG:4326')
        proj_3395 = pyproj.CRS('EPSG:3395')
        project = pyproj.Transformer.from_crs(proj_4326, proj_3395, always_xy=True).transform

        geometries = []
        geometries_coordinates = []

        ids = []
        ids_coordinates = []
        counter_id_coordinates = 0

        # triangulate
        mesh = []
        for id, poly in enumerate(polygons):

            ids.append(id)

            nodes = []
            rings = []

            # outer
            nodes = poly[0]
            rings.append(len(nodes))

            inner_lng_lat = []

            # inner
            for inner in poly[1]:
                inner_lng_lat.append([(elem[1], elem[0]) for elem in inner])
                nodes.extend(inner)
                rings.append(len(nodes))
            
            outer_lng_lat = [(elem[1], elem[0]) for elem in poly[0]]

            geometries.append(transform(project,Polygon(outer_lng_lat, inner_lng_lat)))

            nodes = np.array(nodes)

            indices = earcut.triangulate_float64(nodes, rings)

            indices = np.flip(indices, axis=0)

            # empty triangulation
            if(len(indices) == 0 or (len(indices) % 3) > 0):
                raise errors.InvalidPolygon('Invalid triangulation')

            # bad triangulation
            nodes = nodes.flatten().tolist()
            indices = indices.tolist()
            dev = utils.deviation(nodes, rings, 2, indices)
            if(abs(dev) > 0.001):
                raise errors.InvalidPolygon('Invalid deviation (%f)'%dev)
            
            nodes = utils.convert_projections("4326", "3395", nodes)

            for i in range(int(len(nodes)/2)):
                geometries_coordinates.append(Point(nodes[i*2], nodes[i*2+1]))
                ids_coordinates.append(counter_id_coordinates)
                counter_id_coordinates += 1

            if convert2dto3d:
                nodes = utils.from_2d_to_3d(nodes)

            mesh.append({'type': 'type', 'geometry': {'coordinates': [round(item,4) for item in nodes], 'indices': indices}})
        
        gdf = gpd.GeoDataFrame({'geometry': geometries, 'id': ids}, crs=3395)

        gdf_coordinates = gpd.GeoDataFrame({'geometry': geometries_coordinates, 'id': ids_coordinates}, crs=3395)

        return {'data': mesh, 'gdf': {'objects': gdf, 'coordinates': gdf_coordinates, 'coordinates3d': None}}

    def parse_osm(osm_json):
        '''
            Parses the OSM data into ways and multiways

            Args:
                osm_json (object): OSM data recovered from the API

            Returns:
                parsed_data (object): Object containing ways and multiways
        '''

        ways = {}
        multiways = {}

        for el in osm_json['elements']:
            if el['type']=='way':
                bbox = [el['bounds']['minlat'],el['bounds']['minlon'],el['bounds']['maxlat'],el['bounds']['maxlon']]
                geom = [(x['lat'],x['lon']) for x in el['geometry'] if x != None]
                ways[el['id']] = {'geometry': geom, 'bbox': bbox, 'tags': el['tags']}
            elif el['type']=='relation':
                multiways[el['id']] = []
                for ell in el['members']:
                    if ell['type']=='way':
                        role = ell['role']
                        if role == 'outer':
                            last = len(multiways[el['id']])-1
                            if last == -1:
                                multiways[el['id']].append({'outer': [], 'inner': []})
                            bbox = [el['bounds']['minlat'],el['bounds']['minlon'],el['bounds']['maxlat'],el['bounds']['maxlon']]
                            geom = [(x['lat'],x['lon']) for x in ell['geometry'] if x != None]
                            multiways[el['id']][last]['outer'].append({'geometry': geom, 'bbox': bbox, 'tags': el['tags']})
                        elif role == 'inner':
                            last = len(multiways[el['id']])-1
                            if last == -1:
                                multiways[el['id']].append({'outer': [], 'inner': []})
                            bbox = [el['bounds']['minlat'],el['bounds']['minlon'],el['bounds']['maxlat'],el['bounds']['maxlon']]
                            geom = [(x['lat'],x['lon']) for x in ell['geometry'] if x != None]
                            multiways[el['id']][last]['inner'].append({'geometry': geom, 'bbox': bbox, 'tags': el['tags']})
    
        return {'ways': ways, 'multiways': multiways}

    def format_osmium(ways_elements, areas):
        ways = {}
        multiways = {}

        for el in ways_elements['elements']:
            if el['type']=='way':
                bbox = [el['bounds']['minlat'],el['bounds']['minlon'],el['bounds']['maxlat'],el['bounds']['maxlon']]
                geom = [(x['lat'],x['lon']) for x in el['geometry'] if x != None]
                ways[el['id']] = {'geometry': geom, 'bbox': bbox, 'tags': el['tags']}

        for area_id in areas:

            multiways[area_id] = [{
                'outer': areas[area_id]['outer'],
                'inner': areas[area_id]['inner']
            }]

        return {'ways': ways, 'multiways': multiways}
        
    def build_osm_query(coordinates, output_type, bbox=False, layers=['parks', 'water', 'roads','building']):
        '''
            Create an OSM query from layer types

            Args:
                bpoly (float[]): coordinates of bounding polygon
                bbox: if bpoly follows the format [minLat, minLng, maxLat, maxLng]
                output_type (string): Type of output (i.e. geom)
                layers (string[]): List of strings containing the names of the layers that will be included in the OSM query. Possible values: ('coastline', 'parks', 'water', 'roads','building')
                    (default is ['parks', 'water', 'roads','building'])

            Returns:
                query (string): A string representing the OSM query

        '''

        # test_bbox = [40.7065227, -73.9957237, 40.6981960, -74.0112591, 40.7004079, -74.0234041, 40.7110434, -74.0229750, 40.7229778, -74.0183830, 40.7251455, -74.0139055, 40.7065125, -73.9955807]
        # test_bbox = [40.6979579, -74.0146923, 40.7047885, -74.0242195, 40.7178622, -74.0213013, 40.7406970, -74.0177536, 
        #              40.7582171, -74.0124464, 40.7797333, -73.9968252, 40.8037116, -73.9801311, 40.8264477, -73.9618492, 
        #              40.8544337, -73.9488029, 40.8692525, -73.9410782, 40.8791798, -73.9281607, 40.8780249, -73.9232719,
        #              40.8761352, -73.9201283, 40.8747078, -73.9149141, 40.8729788, -73.9099646, 40.8673246, -73.9106727,
        #              40.8627411, -73.9148355, 40.8584904, -73.9190197, 40.8535037, -73.9229608, 40.8492956, -73.9259648,
        #              40.8410680, -73.9312148, 40.8336130, -73.9343619, 40.8267696, -73.9338899, 40.8145508, -73.9331245,
        #              40.8089873, -73.9335251, 40.8032547, -73.9291906, 40.7998264, -73.9279675, 40.7948624, -73.9286399,
        #              40.7887648, -73.9353490, 40.7811608, -73.9405417, 40.7767898, -73.9412713, 40.7703865, -73.9454985,
        #              40.7630899, -73.9517713, 40.7566336, -73.9579082, 40.7433268, -73.9699817, 40.7328711, -73.9716983,
        #              40.7211009, -73.9701891, 40.7088030, -73.9768267, 40.7061541, -73.9964533, 40.7011167, -74.0048647,
        #              40.6980209, -74.0144205
        #             ]

        if(bbox):
            bpoly = ','.join(format(coord,".8f") for coord in coordinates)
        else:
            bpoly = ' '.join(format(coord,".8f") for coord in coordinates)
        query = '[out:json];('
        for layer in layers:
            filters = OSM.get_overpass_filters(layer)
            for ffilter in filters['way']:
                if(bbox):
                    query += 'way%s(%s);'%(ffilter,bpoly)
                else:
                    query += 'way%s(poly:"%s");'%(ffilter,bpoly)
            for ffilter in filters['rel']:
                if(bbox):
                    query += 'rel%s(%s);'%(ffilter,bpoly)
                else:
                    query += 'rel%s(poly:"%s");'%(ffilter,bpoly)
            for ffilter in filters['node']:
                if(bbox):
                    query += 'node%s(%s);'%(ffilter,bpoly)
                else:
                    query += 'node%s(poly:"%s");'%(ffilter,bpoly)


        query += ');out %s;'%output_type

        return query

    def get_overpass_filters(layer_type):

        '''
            Create an osm filter based on urban-toolkit layer types (building, road, coastline, water, parks)

            Args:
                layer_type (string): Name of the layer to generate filters for. Possible values: 'building', 'road', 'coastline', 'water', 'parks'

            Returns:
                result (object): An object containing all the filters for way, rel
        '''

        filters = {}
        filters['way'] = []
        filters['node'] = []
        filters['rel'] = []

        if layer_type == 'water':
            natural_types = ['water', 'wetland', 'bay', 'strait', 'spring']
            water_types = ['pond', 'reservoir', 'lagoon', 'stream_pool', 'lake', 'pool', 'canal', 'river']
            
            filters['way'].extend(['["natural"="%s"]'%t for t in natural_types])
            filters['rel'].extend(['["natural"="%s"]'%t for t in natural_types])
            filters['way'].extend(['["water"="%s"]'%t for t in water_types])
            filters['rel'].extend(['["water"="%s"]'%t for t in water_types])

        if layer_type == 'parks':
            natural_types = ['wood', 'grass']
            land_use_types = ['wood', 'grass', 'forest', 'orchad', 'village_green', 'vineyard', 'cemetery', 'meadow', 'village_green']
            leisure_types = ['dog_park', 'park', 'playground', 'recreation_ground']

            filters['way'].extend(['["natural"="%s"]'%t for t in natural_types])
            filters['rel'].extend(['["natural"="%s"]'%t for t in natural_types])
            filters['way'].extend(['["landuse"="%s"]'%t for t in land_use_types])
            filters['rel'].extend(['["landuse"="%s"]'%t for t in land_use_types])
            filters['way'].extend(['["leisure"="%s"]'%t for t in leisure_types])
            filters['rel'].extend(['["leisure"="%s"]'%t for t in leisure_types])

        if layer_type == 'roads':
            filters['way'].extend(['["highway"]["area"!~"yes"]["highway"!~"cycleway|footway|proposed|construction|abandoned|platform|raceway"]'])
            # filters['way'].extend(['"area"!~"yes"'])
            # filters['way'].extend(['"highway"!~"proposed|construction|abandoned|platform|raceway)"'])
            # filters['rel'].extend(['"highway"="%s"'%t for t in highway_types])

        if layer_type == 'coastline':
            filters['way'].extend(['["natural"="coastline"]'])

        if layer_type == 'buildings':
            filters['way'].extend(['["building"]'])
            filters['way'].extend(['["building:part"]'])
            filters['way'].extend(['["type"="building"]'])

            # filters['way'].extend(['["building:levels"]'])
            # filters['way'].extend(['["building:min_level"]'])
            # filters['way'].extend(['["building:height"]'])
            # filters['way'].extend(['["roof:levels"]'])
            # filters['way'].extend(['["roof:height"]'])
            # filters['way'].extend(['["roof:shape"]'])
            # filters['rel'].extend(['["building"]'])
            # filters['rel'].extend(['["building:levels"]'])
            # filters['rel'].extend(['["building:min_level"]'])
            # filters['rel'].extend(['["building:height"]'])
            # filters['rel'].extend(['["roof:levels"]'])
            # filters['rel'].extend(['["roof:height"]'])
            # filters['rel'].extend(['["roof:shape"]'])

        return filters

    def get_osmium_filters(layer_type):
        
        '''
            Create OSM filters for osmium based on urban-toolkit layer types (building, road, coastline, water, parks)

            Args:
                layer_type (string): Name of the layer to generate filters for. Possible values: 'building', 'road', 'coastline', 'water', 'parks'

            Returns:
                result (object): An object containing all the filters for way, rel
        '''

        filters = {}
        filters['way'] = {'disqualifiers': {}}
        filters['rel'] = {'disqualifiers': {}}

        if layer_type == 'water':
            natural_types = ['water', 'wetland', 'bay', 'strait', 'spring']
            water_types = ['pond', 'reservoir', 'lagoon', 'stream_pool', 'lake', 'pool', 'canal', 'river']

            filters['way']['natural'] = natural_types
            filters['rel']['natural'] = natural_types

            filters['way']['water'] = water_types
            filters['rel']['water'] = water_types

        if layer_type == 'parks':
            natural_types = ['wood', 'grass']
            land_use_types = ['wood', 'grass', 'forest', 'orchad', 'village_green', 'vineyard', 'cemetery', 'meadow', 'village_green']
            leisure_types = ['dog_park', 'park', 'playground', 'recreation_ground']

            filters['way']['natural'] = natural_types
            filters['rel']['natural'] = natural_types

            filters['way']['landuse'] = land_use_types
            filters['rel']['landuse'] = land_use_types

            filters['way']['leisure'] = leisure_types
            filters['rel']['leisure'] = leisure_types

        if layer_type == 'roads':

            filters['way']['highway'] = [-1] # including all highways

            filters['way']['disqualifiers']['area'] = ['yes']
            filters['way']['disqualifiers']['highway'] = ['cycleway', 'footway', 'proposed', 'construction', 'abandoned', 'platform', 'raceway'] # excluding certain types of highway

        if layer_type == 'coastline':
            filters['way']['natural'] = ['coastline']

        if layer_type == 'buildings':
            filters['way']['building'] = [-1]
            filters['way']['building:part'] = [-1]
            filters['way']['type'] = ['building']

        return filters

    def fill_relation_geom_osmium(ways_elements, relation_elements, ways_position, relations_position, areas):

        def update_bounds(way_bounds, bounds):

            if(bounds['minlat'] == None):
                bounds['minlat'] = way_bounds['minlat']
            elif(way_bounds['minlat'] < bounds['minlat']):
                    bounds['minlat'] = way_bounds['minlat']

            if(bounds['minlon'] == None):
                bounds['minlon'] = way_bounds['minlon']
            elif(way_bounds['minlon'] < bounds['minlon']):
                    bounds['minlon'] = way_bounds['minlon']

            if(bounds['maxlat'] == None):
                bounds['maxlat'] = way_bounds['maxlat']
            elif(way_bounds['maxlat'] > bounds['maxlat']):
                    bounds['maxlat'] = way_bounds['maxlat']

            if(bounds['maxlon'] == None):
                bounds['maxlon'] = way_bounds['maxlon']
            elif(way_bounds['maxlon'] > bounds['maxlon']):
                    bounds['maxlon'] = way_bounds['maxlon']
            
            return bounds

        relations = {'elements': []}

        for relation_id in relations_position:
            relation = relation_elements['elements'][relations_position[relation_id]]

            bounds = {
                'minlat': None,
                'minlon': None,
                'maxlat': None, 
                'maxlon': None
            }

            filtered_members = [] # only keeping the members that were included in the bbox

            for member in relation['members']:

                if(member['type'] == 'way' and member['id'] in ways_position): # TODO: verify why some ways that compose certain relations do not appear in the ways array
                # if(member['type'] == 'way' and relation['id'] in areas):
                    way = ways_elements['elements'][ways_position[member['id']]]

                    update_bounds(way['bounds'], bounds)

                    member['geometry'] = way['geometry'].copy()

                if(len(member['geometry']) > 0):
                    filtered_members.append(member)

            relation['members'] = filtered_members
            relation['bounds'] = bounds

            relations['elements'].append(relation)

        return relations


    def discretize_surface_mesh(coords, size=-1):
        poly = Polygon(coords[:,:2])

        # coordinates, indices, ids, _, _, _, _ = Buildings.get_roof(poly, None, 0, size)
        coordinates, indices, ids, _ = Buildings.get_roof(poly, None, 0, size)

        vmesh = vedo.Mesh([coordinates, indices])
        normals = vmesh.normals(cells=False)

        return coordinates, indices, ids, normals

    def create_surface_mesh(bpoly, bbox, nCells = -1, sizeCells = -1):
        '''
            Creates the surface mesh that covers the bounding box

            Args:
                bpoly (float[]): coordinates of bounding polygon
                bbox: if bpoly follows the format [minLat, minLng, maxLat, maxLng]
                sizeCells (number): if sizeCells is not specified and bbox is not a square the size of the cells will be defined according to the biggest side
                nCells (number): infered from sizeCells if not specified
            Returns:
                mesh (object): A json object describing the geometry of the layer
        '''
        
        if(nCells != -1 and sizeCells != -1):
            raise Exception("It is only possible to specify nCells or sizeCells not both")

        if(bbox):
            nodes = [bpoly[0],bpoly[1], bpoly[2],bpoly[1], bpoly[2],bpoly[3], bpoly[0],bpoly[3]]
        else:
            # TODO: crop the bbox according to the polygon
            boundaries = utils.polygon_bpoly(bpoly, bbox).bounds
            nodes = [boundaries[0],boundaries[1], boundaries[2],boundaries[1], boundaries[2],boundaries[3], boundaries[0],boundaries[3]]

        nodes = utils.convert_projections("4326", "3395", nodes)

        gdf = gpd.GeoDataFrame({'geometry': [Polygon([(nodes[0], nodes[1]), (nodes[2], nodes[3]), (nodes[4], nodes[5]), (nodes[6], nodes[7])])], "id": [0]}, crs=3395)

        nodes = utils.from_2d_to_3d(nodes)

        indices = [0, 3, 2, 2, 1, 0]

        geometry = [{'type': 'surface', 'geometry': {'coordinates': nodes, 'indices': indices}}]

        flat_coordinates = geometry[0]['geometry']['coordinates']
        grouped_coordinates = np.reshape(np.array(flat_coordinates), (int(len(flat_coordinates)/3), -1))

        finalSize = sizeCells

        if(finalSize == -1 and nCells != -1):
            finalSize = max(abs(bbox[2] - bbox[0])/nCells, abs(bbox[3] - bbox[1])/nCells)

        coordinates, indices, ids, normals = OSM.discretize_surface_mesh(grouped_coordinates, finalSize)

        geometries_coordinates = []
        ids_coordinates = []

        flat_discretized_coords = [float(elem) for sublist in coordinates for elem in sublist]
        flat_discretized_indices = [int(elem) for sublist in indices for elem in sublist]
        flat_discretized_normals = [float(elem) for sublist in normals for elem in sublist]

        for i in range(int(len(flat_discretized_coords)/3)):
            geometries_coordinates.append(Point(float(flat_discretized_coords[i*3]), float(flat_discretized_coords[i*3+1])))
            ids_coordinates.append(i)

        gdf_coordinates = gpd.GeoDataFrame({'geometry': geometries_coordinates, "id": ids_coordinates}, crs=3395)

        mesh = [{
            'geometry': {
                'coordinates': [float(elem) for sublist in coordinates for elem in sublist],
                'indices': [int(elem) for sublist in indices for elem in sublist],
                'ids': [int(elem) for elem in ids],
                'normals': [float(elem) for sublist in normals for elem in sublist],
                "discardFuncInterval": [0, 0.01]
            }
        }]

        # Temp code to create one object per cell
        # ===========================================================================================

        # mesh = []
        # maxId = max(ids)

        # for i in range(maxId+1):
        #     mesh.append({
        #         'geometry': {
        #             'coordinates': [],
        #             'indices': [],
        #             'ids': [],
        #             'normals': [],
        #             'discardFuncInterval': [0, 0.01]
        #         }
        #     })            

        # coordinates_index_translation = {}

        # # each position of ids indicates the cell to which this triangle belong
        # for id_index, id_content in enumerate(ids):

        #     # adding the coordinates of this triangle
        #     if indices[id_index][0] not in mesh[id_content]['geometry']['indices']:

        #         mesh[id_content]['geometry']['coordinates'].append(float(coordinates[indices[id_index][0]][0]))
        #         mesh[id_content]['geometry']['coordinates'].append(float(coordinates[indices[id_index][0]][1]))
        #         mesh[id_content]['geometry']['coordinates'].append(float(coordinates[indices[id_index][0]][2]))

        #         if indices[id_index][0] not in coordinates_index_translation:
        #             coordinates_index_translation[indices[id_index][0]] = len(mesh[id_content]['geometry']['coordinates'])/3 - 1

        #     if indices[id_index][1] not in mesh[id_content]['geometry']['indices']:

        #         mesh[id_content]['geometry']['coordinates'].append(float(coordinates[indices[id_index][1]][0]))
        #         mesh[id_content]['geometry']['coordinates'].append(float(coordinates[indices[id_index][1]][1]))
        #         mesh[id_content]['geometry']['coordinates'].append(float(coordinates[indices[id_index][1]][2]))

        #         if indices[id_index][1] not in coordinates_index_translation:
        #             coordinates_index_translation[indices[id_index][1]] = len(mesh[id_content]['geometry']['coordinates'])/3 - 1

        #     if indices[id_index][2] not in mesh[id_content]['geometry']['indices']:

        #         mesh[id_content]['geometry']['coordinates'].append(float(coordinates[indices[id_index][2]][0]))
        #         mesh[id_content]['geometry']['coordinates'].append(float(coordinates[indices[id_index][2]][1]))
        #         mesh[id_content]['geometry']['coordinates'].append(float(coordinates[indices[id_index][2]][2]))

        #         if indices[id_index][2] not in coordinates_index_translation:
        #             coordinates_index_translation[indices[id_index][2]] = len(mesh[id_content]['geometry']['coordinates'])/3 - 1

        #     # adding the triangle
        #     mesh[id_content]['geometry']['indices'].append(int(indices[id_index][0]))
        #     mesh[id_content]['geometry']['indices'].append(int(indices[id_index][1]))
        #     mesh[id_content]['geometry']['indices'].append(int(indices[id_index][2]))

        #     #adding the normals of the coordinates
        #     mesh[id_content]['geometry']['normals'].append(float(normals[indices[id_index][0]][0]))
        #     mesh[id_content]['geometry']['normals'].append(float(normals[indices[id_index][0]][1]))
        #     mesh[id_content]['geometry']['normals'].append(float(normals[indices[id_index][0]][2]))

        #     mesh[id_content]['geometry']['normals'].append(float(normals[indices[id_index][1]][0]))
        #     mesh[id_content]['geometry']['normals'].append(float(normals[indices[id_index][1]][1]))
        #     mesh[id_content]['geometry']['normals'].append(float(normals[indices[id_index][1]][2]))

        #     mesh[id_content]['geometry']['normals'].append(float(normals[indices[id_index][2]][0]))
        #     mesh[id_content]['geometry']['normals'].append(float(normals[indices[id_index][2]][1]))
        #     mesh[id_content]['geometry']['normals'].append(float(normals[indices[id_index][2]][2]))

        #     # adding the id of the triangle
        #     mesh[id_content]['geometry']['ids'].append(int(ids[id_index]))

        # for element in mesh:

        #     element['geometry']['indices'] = [int(coordinates_index_translation[value]) for value in element['geometry']['indices']]

        #     # minIndex = min(element['geometry']['indices'])
        #     # element['geometry']['indices'] = [value - minIndex for value in element['geometry']['indices']]

        #     minIds = min(element['geometry']['ids'])
        #     element['geometry']['ids'] = [value - minIds for value in element['geometry']['ids']]

        # ===========================================================================================


        return {'data': mesh, 'gdf': {'objects': gdf, 'coordinates': gdf_coordinates, 'coordinates3d': None}}



import os
import hashlib
import json

def _hash_query(query):
    return hashlib.md5(query.encode("utf-8")).hexdigest()

def _save_osm_to_cache(query, response):
    filename = _hash_query(query)
    cachepath = './urbantk_cache/'
    if not os.path.exists(cachepath):
        os.makedirs(cachepath)
    response_str = str(json.dumps(response))
    with open(cachepath+filename, "w", encoding="utf-8") as cache_file:
        cache_file.write(response_str)


def _load_osm_from_cache(query):
    filename = _hash_query(query)
    cachepath = './urbantk_cache/'
    if os.path.isfile(cachepath+filename):
        with open(cachepath+filename, "r", encoding="utf-8") as cache_file:
            response = json.load(cache_file)
            return response
    else:
        return None





import os

from IPython.display import HTML

def get_html(map_id, width, height):
    filepath = os.path.dirname(os.path.realpath(__file__))
    # urbanmaptk = open(filepath+'/ts/dist/urbantkmap.iife.js',mode='r').read()
    print(filepath)
    urbanmaptk = open('../../../dist/build/bundle.min.js',mode='r').read() # load urbantk-react bundle

    width = str(width)
    height = str(height)
    if ('px' not in width) and ('%' not in width):
        width = width+'px'
    if ('px' not in height) and ('%' not in height):
        height = height+'px'

    html = """
        <!DOCTYPE html>
        <html lang="en">
        <head>
        <meta charset="UTF-8">
        <title>urbantkmap widget</title>
        </head>

        <style>
        #{mapid} {{
            width: {width};
            height: {height};
            margin: 0px;

            overflow: hidden;
        }}
        </style>

        <body>
        <div id="root"></div>
        <script>
            {urbanmaptk}
        </script>
        <script>
            var el = document.getElementById('{mapid}');
            var comm_manager = Jupyter.notebook.kernel.comm_manager;
            comm_manager.register_target('{mapid}_initMapView', function (comm, data) {{
               window.comm_handle = comm;
               comm.on_msg(function (data) {{ console.log(data['content']['data']); window.data = data['content']['data']; window.JupyterReact.init('#root', window.data);}});
            }});
        </script>
        
        </body>
        </html>
    """.format(width=width, height=height, urbanmaptk=urbanmaptk, mapid=map_id)

    return HTML(html)


import warnings
warnings.simplefilter("ignore", UserWarning)

import os
import sys
import time
import argparse
import json
import psutil
import threading
import requests, zipfile, io
from flask import Flask, request, send_from_directory, abort, jsonify
from geopy.geocoders import Nominatim
from watchdog.observers import Observer
from watchdog.events import LoggingEventHandler

from utk.utils import *
from utk.files_interface import *

## Carolina ##
## Change import path if moving src/bin files to src/utk ##
from bin.structure import Structure

app = Flask(__name__)
geolocator = Nominatim(user_agent="urbantk")
workdir = './data/'
grammarpath = './data/grammar.json'
bundlepath = './utk-app/'
tspath = './utk-ts/'
frontpath = './utk-frontend/'
address = 'localhost'
port = 5001

@app.after_request
def add_cors_headers(response):
    response.headers.add('Access-Control-Allow-Origin', '*')
    response.headers.add('Access-Control-Allow-Headers', 'Content-Type,Authorization')
    response.headers.add('Access-Control-Allow-Methods', 'GET,PUT,POST,DELETE')
    return response

@app.route('/')
def root():

    if("grammar" in request.args):

        grammarBool = request.args.get('grammar')

        grammar = {}
        with open(grammarpath, "r", encoding="utf-8") as f:
            grammar = json.load(f)

        if(grammarBool == "True"): # activate grammar
            grammar["grammar"] = True            
        elif(grammarBool == "False"):
            grammar["grammar"] = False

        with open(grammarpath, "w", encoding="utf-8") as f:
            f.write(json.dumps(grammar, indent=4))

    return send_from_directory(bundlepath, 'index.html')

@app.route('/<path:name>')
def serve_bundle(name):

    # Replace with correct 
    if name == 'main.js':
        with open(os.path.join(bundlepath,name), "r", encoding="utf-8") as f:
            print('%s:%d'%(address,port))
            text = f.read()
            replaced = text.replace('localhost:5001','%s:%d'%(address,port))
            response = app.response_class(
                response=replaced,
                status=200,
                mimetype='text/javascript'
            )
            return response

    return send_from_directory(bundlepath, name)

@app.route('/linkLayers', methods=['GET'])
def serve_linkLayers():

    if("in" not in request.args or "spatial_relation" not in request.args or "out" not in request.args or "outLevel" not in request.args or "inLevel" not in request.args or "abstract" not in request.args):
        abort(400, "Missing one or more parameters of: in, spatial_relation, out, outLevel, inLevel, abstract")

    spatial_relation = request.args.get('spatial_relation')
    out = request.args.get('out')
    inData = request.args.get('in')
    outLevel = request.args.get('outLevel')
    inLevel = request.args.get('inLevel')
    abstract = request.args.get('abstract')

    if(abstract == "true"):
        abstract = True
    else:
        abstract = False

    operation = 'avg'

    if("operation" in request.args):
        operation = request.args.get('operation')

    maxDistance = None

    if("maxDistance" in request.args):
        maxDistance = float(request.args.get('maxDistance'))

    if(maxDistance != None and spatial_relation.upper() != 'NEAREST'):
        abort(400, "Max distance can only be used with the NEAREST spatial_relation")

    defaultValue = 0

    if("defaultValue" in request.args):
        defaultValue = float(request.args.get('defaultValue'))

    fi = FilesInterface()

    fi.setWorkDir(workdir)

    if(fi.existsJoin(out, inData, spatial_relation.upper(), outLevel.upper(), inLevel.upper(), abstract)):
        return ''

    fi.addLayerFromJsonFile(os.path.join(workdir, out+".json"))
    fi.addLayerFromJsonFile(os.path.join(workdir, inData+".json"), abstract=abstract)

    if(abstract):
        if(maxDistance != None):
            fi.attachAbstractToPhysical(out, inData, outLevel, inLevel, spatial_relation, operation, maxDistance, default_value=defaultValue)
        else:
            fi.attachAbstractToPhysical(out, inData, outLevel, inLevel, spatial_relation, operation, default_value=defaultValue)
    else:
        if(maxDistance != None):
            fi.attachPhysicalLayers(out, inData, spatial_relation, outLevel, inLevel, maxDistance, default_value=defaultValue)
        else:
            fi.attachPhysicalLayers(out, inData, spatial_relation, outLevel, inLevel, default_value=defaultValue)

    fi.saveJoined(workdir)

    return ''

@app.route('/clearLinks', methods=['GET'])
def serve_clearLinks():

    if("layer" not in request.args):
        print("cleaning joined layers of all layers")
    else:
        print("clearning joined layers of "+request.args.get("layer"))
    
    return ''

@app.route('/files/<path:path>')
def serve_files(path):
    return send_from_directory(workdir, path)

@app.route('/getGrammar', methods=['GET'])
def serve_getGrammar():
    grammar = {}
    with open(grammarpath, "r", encoding="utf-8") as f:
        grammar = json.load(f)

    response = app.response_class(
        response=json.dumps(grammar),
        status=200,
        mimetype='application/json'
    )
    return response

@app.route('/getLayer', methods=['GET'])
def serve_getLayer():

    if("layer" not in request.args):
        abort(400, "Missing one or more parameters of: layer")

    layer = request.args.get('layer')

    layer_json = {}

    with open(os.path.join(workdir,layer+".json"), "r", encoding="utf-8") as f:
        layer_json = json.load(f)

    return json.dumps(layer_json, indent=4)

@app.route('/solveNominatim', methods=['GET'])
def serve_solveNominatim():

    if("text" not in request.args):
        abort(400, "Missing one or more parameters of: text")

    text = request.args.get('text')

    location = geolocator.geocode(text, timeout=5)

    convertedProj = utk.utils.convert_projections("4326", "3395", [location.latitude, location.longitude])

    return json.dumps({
        'position': convertedProj+[3], 
        'direction': {
            'right': [0,0,3000],
            'lookAt': [0,0,0],
            'up': [0,1,0]
        }
    })

# @app.route('/addRenderStyles', methods=['GET'])
# def serve_addRenderStyles():

#     grammar = {}

#     with open(os.path.join(workDir,"grammar.json"), "r", encoding="utf-8") as f:
#         grammar = json.load(f)

#     layersInfo = {}

#     for knot in grammar["components"][0]["knots"]:
#         if('knotOp' not in knot or knot['knotOp'] != True):
            
#             for index, link in enumerate(knot['integration_scheme']):

#                 layer = link['out']
#                 buildings = False
#                 triangles = False
#                 interactions = False
#                 embeddedPlots = False
#                 abstract = 'in' in link
#                 data = {}

#                 if(layer not in layersInfo):
#                     with open(os.path.join(workDir,layer+".json"), "r", encoding="utf-8") as f:
#                         data = json.load(f)
#                 else:
#                     data = layersInfo[layer]['data']

#                 if(data["type"] == "TRIANGLES_3D_LAYER" or data["type"] == "TRIANGLES_3D_LAYER"):
#                     triangles = True

#                 if(data["type"] == "BUILDINGS_LAYER"):
#                     buildings = True

#                 for i in range(len(grammar["components"][0]['map']["knots"])):
#                     if(grammar["components"][0]['map']["knots"][i] == knot["id"] and grammar["components"][0]["map"]["interactions"][i] != "NONE"):
#                         if(index == len(knot['integration_scheme'])-1): # only the layers that will be rendered can be interacted with
#                             interactions = True
#                         break

#                 for i in range(len(grammar["components"][0]["plots"])):
#                     if(knot["id"] in grammar["components"][0]["plots"][i]["knots"] and grammar["components"][0]["plots"][i]["arrangement"] == "SUR_EMBEDDED" or grammar["components"][0]["plots"][i]["arrangement"] == "FOOT_EMBEDDED"):
#                         embeddedPlots = True
#                         break

#                 if(layer not in layersInfo):
#                     layersInfo[layer] = {
#                         "layer": layer,
#                         "triangles": triangles,
#                         "buildings": buildings,
#                         "interactions": interactions,
#                         "embeddedPlots": embeddedPlots,
#                         "abstract": abstract,
#                         "data": data
#                     }
#                 else:
#                     layersInfo[layer]['triangles'] =  layersInfo[layer]['triangles'] or triangles
#                     layersInfo[layer]['buildings'] =  layersInfo[layer]['buildings'] or buildings
#                     layersInfo[layer]['interactions'] =  layersInfo[layer]['interactions'] or interactions
#                     layersInfo[layer]['embeddedPlots'] =  layersInfo[layer]['embeddedPlots'] or embeddedPlots
#                     layersInfo[layer]['abstract'] =  layersInfo[layer]['abstract'] or abstract

#     for layer in layersInfo:
#         renderStyles = []

#         # coloring shader
#         if(not layersInfo[layer]['abstract']):
#             renderStyles.append("SMOOTH_COLOR")
#         elif(not layersInfo[layer]['buildings']):
#             renderStyles.append("SMOOTH_COLOR_MAP")
#         else:
#             renderStyles.append("SMOOTH_COLOR_MAP_TEX")

#         if(layersInfo[layer]['interactions']):
#             renderStyles.append("PICKING")

#         if(layersInfo[layer]['buildings'] and layersInfo[layer]['embeddedPlots']):
#             renderStyles.append("ABSTRACT_SURFACES")

#         layersInfo[layer]['data']['renderStyle'] = renderStyles

#         with open(os.path.join(workDir,layer+".json"), "w", encoding="utf-8") as f:
#             f.write(json.dumps(layersInfo[layer]['data']))

#     return ''

@app.route('/writeImpactViewData', methods=['POST'])
def writeImpactViewData():
    
    impactData = request.json['data']

    with open(os.path.join(workDir,"impactView.json"), "w", encoding="utf-8") as f:
        f.write(json.dumps(impactData))

    return ''

@app.route('/updateGrammar', methods=['POST'])
def serve_updateGrammar():

    grammar = request.json['grammar']
    filename = request.json['filename']
    with open(os.path.join(workdir, filename+".json"), "w", encoding="utf-8") as f:
        f.write(json.dumps(grammar, indent=4))

    return ''

def list_used_ports():
    print("Ports used by utk:")
    for p in psutil.process_iter():
        try:
            if 'python' in p.cmdline()[0]:
                if os.path.basename(__file__) in p.cmdline()[1] or 'utk' in p.cmdline()[1]:
                    connections = p.connections()
                    for c in connections:
                        print(c.laddr.port)
        except:
            pass

def stop_used_ports():
    print("Stopping utk:")
    for p in psutil.process_iter():
        try:
            if 'python' in p.cmdline()[0]:
                if os.path.basename(__file__) in p.cmdline()[1] or 'utk' in p.cmdline()[1]:
                    if 'example' in p.cmdline()[2] or 'start' in p.cmdline()[2]:
                        p.kill()
        except:
            pass

def download_example():
    r = requests.get('https://github.com/urban-toolkit/urbantk/raw/master/examples/simple_example.zip')
    z = zipfile.ZipFile(io.BytesIO(r.content))
    z.extractall('.')

def main():
    global workdir
    global bundlepath
    global grammarpath
    global tspath
    global frontpath
    global address
    global port

    parser = argparse.ArgumentParser(description='Urban Toolkit')
    parser.add_argument('mode', nargs=1, choices=['start', 'list', 'stop', 'example'], help='Start, list or stop utk servers, or start server with a simple example.')
    parser.add_argument('-d', '--data', nargs='?', type=str, required=False, default=None, help='Path to data folder.')
    parser.add_argument('-b', '--bundle', nargs='?', type=str, required=False, help='Path to app bundle (defaults to installed utk bundle).')
    parser.add_argument('-g', '--grammar', nargs='?', type=str, required=False, default=None, help='Path to grammar JSON file, if different from [DATA]/grammar.json (default: [DATA]/grammar.json).')
    parser.add_argument('-a', '--address', nargs='?', type=str, required=False, default='localhost', help='Server address (default: %(default)s).')
    parser.add_argument('-p', '--port', nargs=1, type=int, required=False, default='5001', help='Server port (default: %(default)s).')
    parser.add_argument('-w', '--watch', action='store_true', help='Watch folders, and re-build if there are changes.')


    args = parser.parse_args()

    workdir = args.data
    bundlepath = args.bundle
    grammarpath = args.grammar
    address = args.address
    port = args.port
    mode = args.mode[0]

    if mode == 'stop':
        stop_used_ports()
    elif mode == 'list':
        list_used_ports()
    else:
        if mode == 'example':
            download_example()
            workdir = './simple_example/'
            mode = 'start'

        if workdir == None:
            print("Error: --data not specified.")
            exit(1)

        if grammarpath == None:
            grammarpath = os.path.join(workdir,"grammar.json")

        # check if grammar exists
        if os.path.isfile(grammarpath) is False:
            print("Error: %s does not exist, check arguments."%grammarpath)
            exit(1)

        # check if bundle exist
        if bundlepath is None:
            bundlepath = os.path.join(os.path.dirname(__file__), frontpath+'/build/utk-app')
        if os.path.exists(bundlepath) is False:
            print("Error: %s does not exist, check bundle path."%bundlepath)
            exit(1)

        # absolute paths
        workdir = os.path.abspath(workdir)
        bundlepath = os.path.abspath(bundlepath)
        grammarpath = os.path.abspath(grammarpath)
        tspath = os.path.abspath(tspath)
        frontpath = os.path.abspath(frontpath)

        if args.watch:
            # utk-ts observer
            ts_observer = Observer()
            def run_ts():
                class Event(LoggingEventHandler):
                    def dispatch(self, event):
                        os.system('cd %s && npm run build'%(tspath))
                        os.system('cd %s && npm run build'%(frontpath))
                        print("Build done!")
                event_handler = Event()
                ts_observer.schedule(event_handler, tspath+'/src/', recursive=True)
                ts_observer.start()
            ts_thread = threading.Thread(target=run_ts)
            ts_thread.start()

            # utk-frontend observer
            frontend_observer = Observer()
            def run_frontend():
                class Event(LoggingEventHandler):
                    def dispatch(self, event):
                        os.system('cd %s && npm run build'%(frontpath))
                        print("Build done!")
                event_handler = Event()
                frontend_observer.schedule(event_handler, frontpath+'/src/', recursive=True)
                frontend_observer.start()
            frontend_thread = threading.Thread(target=run_frontend)
            frontend_thread.start()

        app.run(host=address, port=port)



###################################################################################
################################ CAROL ############################################
###################################################################################

struct = Structure()

@app.route("/fields", methods=('GET', 'POST'))
def handleFields():
    if request.method == 'GET':

        fields = struct.getFields()
        
        return json.dumps(fields)

@app.route('/temporalData', methods=('GET', 'POST'))
def handleTseries():
    if request.method == 'POST':
        data = request.json

        if "arr" in data.keys():
            fieldKeys = data["arr"]
            temporalData = struct.getTemporalData(fieldKeys)

        return json.dumps(temporalData)

@app.route("/hmap", methods=('POST',))
def handleMap():
    if request.method == 'POST':
        data = request.json

        cat = data['cat']
        fld = data['fld']

        if fld == 'tmp' and cat == 'obs':
            resp = [{"x": 2, "y": 4}, {"x": 5, "y": 8}]
        else:
            resp = [{"x": 45, "y": 87}, {"x": 235, "y": 61}]

        # resp = [{"x": 2, "y": 4}, {"x": 5, "y": 8}] if data == 'tmp' else [{"x": 45, "y": 87}, {"x": 235, "y": 61}]
        
        return json.dumps(resp)

@app.route("/hmat", methods=('POST',))
def handleHmat():
    if request.method == 'POST':
        data = request.json

        pt   = data['pt']
        time = data['time']

        resp = struct.getHmat(pt, time)

        return json.dumps(resp)

@app.route("/scatter", methods=('POST',))
def handleScatter():
    if request.method == 'POST':
        data = request.json

        field1 = data['fld1']
        field2 = data['fld2']

        resp = struct.getScatter(field1, field2)

        return json.dumps(resp)


if __name__ == '__main__':
    main()



import os
import sys
import json
import numpy as np
from netCDF4 import Dataset
from termcolor import colored
# from wrf import getvar, interplevel, to_np, ALL_TIMES
from pyproj import Transformer
# import xarray as xr

# class WRFOutputReader(object):
#     def __init__(self) -> None:
#         self.__className = "WRFOutputReader"
#         self.__variablesList = ('PM10', 'so2', 'co', 'o3', 'PM2_5_DRY', 'no', 'no2', 'no3', 'T2', 'rh', 'wind10m', 'accRain', '3hRain', '24hRain')
#         self.resetClass()
        
#     def __processRain(self, time_idxs, interval):
#         rainc = self.__ncData.variables['RAINC'][:]
#         rainnc = self.__ncData.variables['RAINNC'][:]

#         accRain = rainc + rainnc

#         data = []

#         if not interval:
#             data = accRain if len(time_idxs) == 0 else np.array(accRain[t] for t in time_idxs)

#         else:
#             if len(time_idxs) == 0:
#                 data = [accRain[t] - accRain[t - interval] for t in range(0, self.__nTimes, interval)]

#             else:
#                 for t in time_idxs:
#                     if t - interval >= 0:
#                         data.append(accRain[t] - accRain[t - interval])
            
#             # elif all(t - interval >= 0 for t in time_idxs):
#             #     data = [accRain[t] - accRain[t - interval] for t in time_idxs]
                
#             # else:
#             #     msg = f"[{self.__className} - __processRain() - Negative time step.]"
#             #     raise ValueError(colored(msg, 'red'))
            
#             data = np.array(data)

#         return data

#     def __process4DVariable(self, var_key, level, time_idxs):
#         data = []

#         if len(time_idxs) == 0: time_idxs = range(self.__nTimes)

#         for t in time_idxs:
#             var_matrix = getvar(self.__ncData, var_key, timeidx=t,
#                         squeeze=False, meta=False)
#             pressure_matrix = getvar(self.__ncData, 'pressure', timeidx=t,
#                         squeeze=False, meta=False)

#             var_matrix = to_np(interplevel(var_matrix, pressure_matrix, level))

#             if np.ma.isMaskedArray(var_matrix):
#                 var_matrix = list(var_matrix)
#                 # var_matrix.reverse()
#             data.append(var_matrix)


#         data = np.array(data)
#         return data
    
#     def __processWind10m(self, level, time_idxs):
#         data = []
        
#         if len(time_idxs) == 0: time_idxs = range(self.__nTimes)

#         for t in time_idxs:
#             pressure_matrix = getvar(self.__ncData, 'pressure', timeidx=t,
#                 squeeze=False, meta=False)
#             u10_matrix = getvar(self.__ncData, 'u10', timeidx=t,
#                         squeeze=False, meta=False)
#             v10_matrix = getvar(self.__ncData, 'v10', timeidx=t,
#                         squeeze=False, meta=False)
#             wspd = ''
    
#     def getGridId(self):
#         return self.__gridId
    
#     def getStartDate(self):
#         return self.__startDate
    
#     def getGridDimensions(self):
#         return self.__nLat, self.__nLon

#     def getNTimes(self):
#         return self.__nTimes
    
#     def getVariableData(self, var_key, time_idxs, var_level=None):
#         active_var = None
#         data = []

#         # Set active_var
#         for i in range(len(self.__variablesList)):
#             if var_key.lower() == self.__variablesList[i].lower(): 
#                 active_var = self.__variablesList[i]
#                 break
        
#         # Set data according to the variable
#         if active_var:
#             if 'Rain' in active_var:
#                 interval = None
#                 if active_var == '3hRain':
#                     interval = 3
#                 elif active_var == '24hRain':
#                     interval = 24
#                 self.__processRain(time_idxs, interval)

#             elif active_var == 'T2':
#                 data = self.__ncData.variables[active_var][:] if len(time_idxs) == 0 else np.array([self.__ncData.variables[active_var][t] for t in time_idxs])
#                 data -= 273 # Kelvin to Celsius
            
#             elif active_var == 'wind10m':
#                 # to do
#                 # https://medium.com/the-barometer/plotting-wrf-data-using-python-wrf-python-and-cartopy-edition-b7bf45ff46bb
#                 data = []

#             else:
#                 data = self.__process4DVariable(active_var, var_level, time_idxs)
                    
#         else:
#             sys.exit()
              
#         return data
    
#     def getLatLon(self):
#         return self.__latList, self.__lonList
    
#     def setNcData(self, ncFilePath):
#         try:
#             self.__ncData = Dataset(ncFilePath)

#         except Exception as error:
#             msg = f"[{self.__className} - setNcData]\n{error}"
#             print(colored(f"{msg}", 'red'))
#             sys.exit()

#     def resetClass(self):
#         self.__ncData = None
      
#         self.__latList = None
#         self.__lonList = None

#         self.__nLat = None
#         self.__nLon = None

#         self.__nTimes = None
#         self.__gridId = None
#         self.__startDate = None
    
#     def setAttributes(self):
#         # latMatrix = sorted(self.__ncData.variables['XLAT'][:][0], key=lambda x: x[0], reverse=True)
#         latMatrix = self.__ncData.variables['XLAT'][:][0]
#         lonMatrix = self.__ncData.variables['XLONG'][:][0]
        
#         self.__latList = np.array([row[0] for row in latMatrix])
#         self.__lonList = np.array(lonMatrix[0])

#         self.__nLat = self.__ncData.dimensions['south_north'].size
#         self.__nLon = self.__ncData.dimensions['west_east'].size

#         self.__nTimes = self.__ncData.dimensions['Time'].size

#         self.__gridId = self.__ncData.GRID_ID
#         self.__startDate = self.__ncData.SIMULATION_START_DATE.replace(":", "")

#         # for k in self.__ncData.variables.keys(): print(k)
#         for k in self.__ncData.variables.keys(): print(k)
 
# def thematic_from_netcdf(model, file, file_path, target_variables, coordinates_projection, time_idxs=[], bbox=[]):
#     transformer = Transformer.from_crs(coordinates_projection, 3395)
#     ncReader = None

#     # Set ncReader
#     if model.lower() == 'wrf':
#         ncReader = WRFOutputReader()
    
#     ncReader.setNcData(file_path)
#     ncReader.setAttributes()

#     start_date = ncReader.getStartDate()
#     grid_id = ncReader.getGridId()

#     lat_array, lon_array = ncReader.getLatLon()

#     n_lat, n_lon = ncReader.getGridDimensions()
#     # n_times = ncReader.getNTimes()

#     lat_idxs = range(n_lat)
#     lon_idxs = range(n_lon)
    
#     latmin_idx = 0
#     latmax_idx = n_lat-1
        
#     lonmin_idx = 0
#     lonmax_idx = n_lon-1

#     time_idxs.sort()
    
#     if len(bbox) > 0:

#         latmin, lonmin = bbox[0], bbox[1]
#         latmax, lonmax = bbox[2], bbox[3]

#         lat_idxs = [i for i in range(n_lat) if lat_array[i] >= latmin and lat_array[i] <= latmax]
#         lon_idxs = [j for j in range(n_lon) if lon_array[j] >= lonmin and lon_array[j] <= lonmax]
    
#         latmin_idx = lat_idxs[0]
#         latmax_idx = lat_idxs[len(lat_idxs)-1]
        
#         lonmin_idx = lon_idxs[0]
#         lonmax_idx = lon_idxs[len(lon_idxs)-1]

#     try:
#         points = []

#         for latidx in range(latmin_idx, latmax_idx + 1):
#             for lonidx in range(lonmin_idx, lonmax_idx + 1):
#                 points.append((lat_array[latidx], lon_array[lonidx]))

#         coordinates = []

#         for point in transformer.itransform(points):
#             coordinates.append(float(point[0]))
#             coordinates.append(float(point[1]))
#             coordinates.append(0)

#     except Exception as error:
#             msg = f"[thematic_from_wrf]\n{error}"
#             print(colored(f"{msg}", "red"))
#             sys.exit()
    
    
#     for var_obj in target_variables:
#         var_key   = var_obj.get('variable')
#         var_level = var_obj.get('level')

#         values = []
#         layer_id = f"{file}_{var_key}"
        
#         data = ncReader.getVariableData(var_key, time_idxs, var_level)
        
#         try:
#             for latidx in range(latmin_idx, latmax_idx + 1):
#                 for lonidx in range(lonmin_idx, lonmax_idx + 1):
#                     pt_arr = [round(float(matrix[latidx][lonidx]), 2) for matrix in data]
#                     values.append(pt_arr)
        
#             abstract_json = {
#                 "id": layer_id,
#                 "coordinates": coordinates,
#                 "values": values
#             }

#             json_object = json.dumps(abstract_json)

#             directory = os.path.dirname(filepath)

#             with open(os.path.join(directory,layer_id+".json"), "w") as outfile:
#                 outfile.write(json_object)

#         except Exception as error:
#             msg = f"[thematic_from_wrf]\n{error}"
#             print(colored(f"{msg}", "red"))
#             sys.exit()


if __name__ == '__main__':
    # path = f'../../examples/wrf_chicago'
    # startDate = "2016-07-01"
    # gId = "2"

    # filepath = f"{path}/wrfout_d0{gId}_{startDate}.nc"

    path = f'../../examples/wrf_chicago'
    startDate = "2023-06-09_20_00_00"
    gId = "3"

    file = f"wrfout_d0{gId}_{startDate}"
    filepath = f"{path}/{file}"
    # filepath = f"{path}/wrfout_d03_2023-06-09_20_00_00.nc"

    # obj1 = {'variable': 'T2', 'level': None}
    # obj2 = {'variable': 'rh', 'level': 850}
    # obj3 = {'variable': '3hRain', 'level': None}
    # varList = [obj1, obj2, obj3]

    obj1 = {'variable': 'T2', 'level': None}
    obj2 = {'variable': 'rh', 'level': 850}
    obj3 = {'variable': '3hRain', 'level': None}
    varList = [obj3]

    # thematic_from_netcdf('wrf', file, filepath, varList, 4326)
    # thematic_from_netcdf('wrf', file, filepath, varList, 4326, [t for t in range(1, 5)], [33, -94, 42, -81])
    # thematic_from_netcdf('wrf', file, filepath, varList, 4326, list(range(0, 2)), [33, -94, 42, -81])
    # thematic_from_netcdf('wrf', file, filepath, varList, 4326)
    


import rasterio
import geopandas as gpd
import pandas as pd
import json
import mmap
import zlib
import os
import time
import hashlib
import ast

from shapely import wkt

# Utility Functions
# transforms the whole input into a dict (json) in depth
# def toJsonInput(input):
#     print("here", input)
#     parsedJson = input
#     if(parsedJson['dataType'] == 'outputs'):
#         for key, elem in enumerate(parsedJson['data']):
#             parsedJson['data'][key] = toJsonInput(elem)

#     return parsedJson

# I/O Type Checking
def checkIOType(data, boxType, input=True):
    if input:
        validate_input(data, boxType)
    else:
        validate_output(data, boxType)


# Input Validation
def validate_input(data, boxType):
    if boxType in ['DATA_EXPORT', 'DATA_CLEANING']:
        check_dataframe_input(data, boxType)
    elif boxType == 'DATA_TRANSFORMATION':
        check_transformation_input(data, boxType)

# Output Validation
def validate_output(data, boxType):
    if boxType in ['DATA_LOADING', 'DATA_CLEANING', 'DATA_TRANSFORMATION']:
        check_valid_output(data, boxType)
    elif boxType == 'DATA_EXPORT':
        raise Exception(f'{boxType} does not support output')


# Input Type Checks
def check_dataframe_input(data, boxType):
    if data['dataType'] == 'outputs' and len(data['data']) > 1:
        raise Exception(f'{boxType} only supports one input')

    valid_types = {'dataframe', 'geodataframe'}
    if data['dataType'] == 'outputs':
        for elem in data['data']:
            if elem['dataType'] not in valid_types:
                raise Exception(f'{boxType} only supports DataFrame and GeoDataFrame as input')
    elif data['dataType'] not in valid_types:
        raise Exception(f'{boxType} only supports DataFrame and GeoDataFrame as input')

def check_transformation_input(data, boxType):
    valid_types = {'dataframe', 'geodataframe', 'raster'}
    if data['dataType'] == 'outputs' and len(data['data']) > 2:
        raise Exception(f'{boxType} only supports one or two inputs')

    if data['dataType'] == 'outputs':
        for elem in data['data']:
            if elem['dataType'] not in valid_types:
                raise Exception(f'{boxType} only supports DataFrame, GeoDataFrame, and Raster as input')
    elif data['dataType'] not in valid_types:
        raise Exception(f'{boxType} only supports DataFrame, GeoDataFrame, and Raster as input')

def check_valid_output(data, boxType):
    valid_types = {'dataframe', 'geodataframe', 'raster'}

    if data['dataType'] == 'outputs':
        if len(data['data']) > 1 and boxType != 'DATA_LOADING':
            raise Exception(f'{boxType} only supports one output')

        for elem in data['data']:
            if elem['dataType'] not in valid_types:
                raise Exception(f'{boxType} only supports DataFrame, GeoDataFrame, and Raster as output')

    elif data['dataType'] not in valid_types:
        raise Exception(f'{boxType} only supports DataFrame, GeoDataFrame, and Raster as output')

def save_memory_mapped_file(data, shared_disk_path='./.curio/data/'):
    """
    Saves the input data as a memory-mapped JSON file with a unique name.

    Args:
        input_data (dict): The data to be saved.
        shared_disk_path (str): Path to the directory for saving the file.

    Returns:
        str: The path of the saved memory-mapped file.
    """
    launch_dir = os.environ.get("CURIO_LAUNCH_CWD", os.getcwd())    
    resolved_path = os.path.join(launch_dir, shared_disk_path)
    # Ensure the directory exists
    os.makedirs(resolved_path, exist_ok=True)
    
    json_bytes = json.dumps(data, ensure_ascii=False).encode('utf-8')
    input_hash = hashlib.sha256(json_bytes[:512]).digest()[:4].hex()

    # Create a unique filename using hash of the input and current time
    timestamp = str(int(time.time()))
    unique_filename = f"{timestamp}_{input_hash[:10]}.data" 
    file_path = os.path.join(resolved_path, unique_filename)

    # Save the input data directly without compression as a memory-mapped file
    compressed_data = zlib.compress(json_bytes)
    # compressed_data = json.dump(data, file, ensure_ascii=False) # json.dumps(data).encode('utf-8')
    with open(file_path, "wb") as file:
        file.write(compressed_data)
        file.flush()

    return file_path


def load_memory_mapped_file(file_path):
    """
    Loads the JSON data from the specified memory-mapped JSON file.

    Args:
        file_path (str): The path of the memory-mapped JSON file to load.

    Returns:
        dict: The loaded JSON data.
    """
    if not os.path.exists(file_path):
        raise FileNotFoundError(f"The file {file_path} does not exist.")

    # Using mmap for efficient memory-mapped loading
    with open(file_path, "rb") as file:
        with mmap.mmap(file.fileno(), 0, access=mmap.ACCESS_READ) as mmapped_file:
            # Decompress and decode directly from the memory-mapped file
            decompressed_data = zlib.decompress(mmapped_file[:])
            data = json.loads(decompressed_data.decode('utf-8'))
    return data


def parse_primitive(data_type, data_value):
    try:
        return ast.literal_eval(data_value)
    except (ValueError, SyntaxError):
        return data_value

def parse_list(data_type, data_value):
    values = []
    for elem in data_value:
        if type(elem) == dict and 'dataType' in elem:
            values.append(parseInput(elem))
        else:
            values.append(elem)
    return values

def parse_dataframe(data_value):
    return pd.DataFrame.from_dict(data_value)

def parse_geodataframe(data_value):
    # df = pd.DataFrame.from_dict(data_value)
    # df['geometry'] = df['geometry'].apply(wkt.loads)
    # gdf = gpd.GeoDataFrame(df, geometry='geometry')
    gdf = gpd.GeoDataFrame.from_features(data_value["features"])
    if 'metadata' in data_value and 'name' in data_value['metadata']:
        gdf.metadata = {'name': data_value['metadata']['name']}
    
    return gdf

def parse_raster(data_value):
    return rasterio.open(data_value)

# Parsing Functions
def parseInput(parsed_json):
    # parsed_json = json.loads(input_str)

    data_type = parsed_json.get('dataType')
    data_value = parsed_json.get('data')

    if data_type in ['int', 'float', 'bool', 'dict', 'str']:
        return parse_primitive(data_type, data_value)
    elif data_type == 'list':
        return parse_list(data_type, data_value)
    elif data_type == 'dataframe':
        return parse_dataframe(data_value)
    elif data_type == 'geodataframe':
        return parse_geodataframe(data_value)
    elif data_type == 'raster':
        return parse_raster(data_value)
    elif data_type == 'outputs':
        return tuple(parseInput(elem) for elem in data_value)

    return None

# Output Functions
def parseOutput(output):
    json_output = {'data': '', 'dataType': ''}
    if isinstance(output, (int, float, bool, str)):
        json_output['data'] = output
        json_output['dataType'] = type(output).__name__
    elif isinstance(output, list):
        json_output['data'] = [parseOutput(elem) for elem in output]
        json_output['dataType'] = type(output).__name__
    elif isinstance(output, dict):
        json_output['data'] = output
        json_output['dataType'] = type(output).__name__
    elif isinstance(output, pd.DataFrame) and not isinstance(output, gpd.GeoDataFrame):
        json_output['data'] = output.to_dict(orient='list')
        json_output['dataType'] = 'dataframe'
    elif isinstance(output, gpd.GeoDataFrame):
        # output['geometry'] = output['geometry'].apply(lambda geom: geom.wkt)
        # json_output['data'] = output.to_dict(orient='list')
        json_output['data'] = output.__geo_interface__
        json_output['dataType'] = 'geodataframe'
        if hasattr(output, 'metadata') and 'name' in output.metadata:
            parsed_geojson = json_output['data']
            parsed_geojson['metadata'] = {'name': output.metadata['name']}
            json_output['data'] = parsed_geojson
    elif isinstance(output, rasterio.io.DatasetReader):
        json_output['data'] = output.name
        json_output['dataType'] = 'raster'
    elif isinstance(output, tuple):
        json_output['data'] = [parseOutput(elem) for elem in output]
        json_output['dataType'] = 'outputs'

    return json_output



from flask import Flask
import os
import logging
from logging.handlers import RotatingFileHandler

from utk_curio.sandbox.config import Config as config_class

from utk_curio.sandbox.extensions import cache


# Flask app
app = Flask(__name__)
app.config.from_object(config_class)

# Flask-Caching
cache.init_app(app)

if not app.debug and not app.testing:
    if app.config['LOG_TO_STDOUT']:
        stream_handler = logging.StreamHandler()
        stream_handler.setLevel(logging.INFO)
        app.logger.addHandler(stream_handler)
    else:
        if not os.path.exists('logs'):
            os.mkdir('logs')
        file_handler = RotatingFileHandler('logs/sandbox.log',
                                           maxBytes=10240, backupCount=10)
        file_handler.setFormatter(logging.Formatter(
            '%(asctime)s %(levelname)s: %(message)s '
            '[in %(pathname)s:%(lineno)d]'))
        file_handler.setLevel(logging.INFO)
        app.logger.addHandler(file_handler)

    app.logger.setLevel(logging.INFO)
    app.logger.info('Sandbox startup')


from utk_curio.sandbox.app import api


from pyproj import Transformer
import pandas as pd
import os
import json
import numpy as np
from netCDF4 import Dataset
from pathlib import Path
from .utils import *

import sys
# from wrf import getvar, interplevel, to_np
import math
'''
    Converts a dataframe into an abstract layer
'''
def thematic_from_df(df, output_filepath, latitude_column, longitude_column, coordinates_projection, z_column = None, value_column=None):
    df_lat_lon = df.drop_duplicates(subset=[latitude_column, longitude_column])

    latitude_list = df_lat_lon[latitude_column].tolist()
    longitude_list = df_lat_lon[longitude_column].tolist()
    
    transformer = Transformer.from_crs(coordinates_projection, 3395)
    points = list(zip(latitude_list, longitude_list))

    coordinates = []
    values_list = []

    for index, point in enumerate(transformer.itransform(points)):
    
        z_list = []
        
        if z_column != None:
            z_list = df[(df[latitude_column] == point[0]) & (df[longitude_column] == point[1])][z_column].toList()
        
        z_value = 0

        if(len(z_list) > 0):
            z_value = z_list[index]

        coordinates.append(point[0])
        coordinates.append(point[1])
        coordinates.append(z_value)

        if value_column != None:
            point_values_list = df[(df[latitude_column] == latitude_list[index]) & (df[longitude_column] == longitude_list[index])][value_column].tolist()
        
        else:
            point_values_list = [1] * len(df_lat_lon.index)
        
        values_list.append(point_values_list)
    
    abstract_json = {
        "id": Path(output_filepath).stem,
        "coordinates": coordinates,
        "values": [elem for elem in values_list]
    }

    json_object = json.dumps(abstract_json)

    directory = os.path.dirname(output_filepath)
    if not os.path.exists(directory):
        os.makedirs(directory)
    
    with open(output_filepath, "w") as outfile:
        outfile.write(json_object)

'''
    Converts a csv file into an abstract layer
'''
def thematic_from_csv(filepath, layer_id, latitude_column, longitude_column, coordinates_projection, z_column = None, value_column=None):
    
    df = pd.read_csv(filepath)
    thematic_from_df(df, os.path.join(os.path.dirname(filepath),layer_id+".json"), latitude_column, longitude_column, coordinates_projection, z_column, value_column)

'''
    Converts a NetCDF (e.g. wrf data) file into an abstract layer
'''
# def thematic_from_netcdf(file_path, variables, coords, layer_id, operations=[], time_indexes=[], bbox={}):
#     def cdiff(scalar, axis=0):
#         '''
#         Performs the same as GrADS function cdiff()
#         http://cola.gmu.edu/grads/gadoc/gradfunccdiff.html
#         The scalar quantity must by 2D.
#         The finite differences calculation ignores the borders, where np.nan is returned.
#         '''
#         # Check if 2D
#         dimScalar = np.size(np.shape(scalar))
#         if dimScalar != 2:
#             print(
#                 "Pystuff Error: scalar must have only 2 dimensions, but it has %d." % dimScalar)
#             return

#         # Length of each dimension
#         lendim0 = np.shape(scalar)[0]
#         lendim1 = np.shape(scalar)[1]

#         # Initialize output var
#         out = np.zeros(np.shape(scalar))
#         out.fill(np.nan)

#         # Centered finite differences
#         for x in np.arange(1, lendim0-1):
#             for y in np.arange(1, lendim1-1):
#                 if axis == 0:
#                     out[x, y] = scalar[x+1, y]-scalar[x-1, y]
#                 elif axis == 1:
#                     out[x, y] = scalar[x, y+1]-scalar[x, y-1]
#                 else:
#                     print(
#                         "Pystuff Error: Invalid axis option. Must be either 0 or 1.")
#                     return
#         return out
    
#     def hdivg(_u, _v, _lat, _lon, _r=6.371*(10**6)):
#         '''
#         Calculates the horizontal divergence (du/dx+dv/dy) exactly like GrADS. 
#         lat and lon are 1D arrays.
#         http://cola.gmu.edu/grads/gadoc/gradfunccdiff.html
#         '''
#         latv, lonv = np.meshgrid(_lat, _lon, indexing='ij')

#         # r = 
#         dtr = np.pi/180
#         dudx = cdiff(_u, axis=1)/cdiff(lonv*dtr, axis=1)
#         dvdy = cdiff(_v*np.cos(latv*dtr), axis=0)/cdiff(latv*dtr, axis=0)
#         out = (dudx + dvdy)/(_r*np.cos(latv*dtr))
#         return out
    
#     ncData = Dataset(file_path)
    
#     lat_arr = []
#     lon_arr = []

#     if len(ncData.variables[coords['lat']].shape) == 1:
#         lat_arr = ncData.variables[coords['lat']]
    
#     else:
#         lat_matrix = ncData.variables[coords['lat']][:][0]
#         lat_arr = np.array([row[0] for row in lat_matrix] )

#     if len(ncData.variables[coords['lon']].shape) == 1:
#         lon_arr = ncData.variables[coords['lon']]
    
#     else:
#         lon_arr = ncData.variables[coords['lon']][:][0][0]

#     min_lat_idx = 0
#     max_lat_idx = len(lat_arr)-1
        
#     min_lon_idx = 0
#     max_lon_idx = len(lon_arr)-1

#     if bbox:
#         min_lat_idx = None
#         max_lat_idx = None
            
#         min_lon_idx = None
#         max_lon_idx = None

#         for i in range(len(lat_arr)):
#             if min_lat_idx is None and lat_arr[i] >= bbox['min_lat'] >=bbox['min_lat']:
#                 min_lat_idx = i

#             elif max_lat_idx is None and lat_arr[i] >= bbox['max_lat'] >=bbox['max_lat']:
#                 max_lat_idx = i
#                 break

#         for i in range(len(lon_arr)):
#             if min_lon_idx is None and lon_arr[i] >= bbox['min_lon'] >=bbox['min_lon']:
#                 min_lon_idx = i

#             elif max_lon_idx is None and lon_arr[i] >= bbox['max_lon'] >=bbox['max_lon']:
#                 max_lon_idx = i
#                 break

#     ###########################################################################################

#     data = []

#     for operation in operations:
#         if operation['type'] == 'vector':
#             if operation['dimension'] == 'space':
#                 if operation['function'] == 'sum':
#                     for v in variables:
#                         var_name = v['name']
#                         data = ncData.variables[v['name']][:] if len(data) == 0 else data + ncData.variables[v['name']][:]

#                 elif operation['function'] == 'interpolation':
#                     pressure   = [v for v in variables if v['key'] == 'pressure'][0]
#                     pressure_name = pressure['name']
                    
#                     var_name = [v['name'] for v in variables if v['key'] != 'pressure'][0]

#                     time_idxs = list(range(5))
#                     level = [p['value'] for p in operation['parameters'] if operation['parameters']['name'] == 'level_0'][0]

#                     for tidx in time_idxs:
#                         pressure_matrix = getvar(ncData, pressure_name, timeidx=tidx,
#                                     squeeze=False, meta=False)
                        
#                         var_matrix = getvar(ncData, var_name, timeidx=tidx,
#                         squeeze=False, meta=False)
                        
#                         var_matrix = to_np(interplevel(var_matrix, pressure_matrix, level))

#                         if np.ma.isMaskedArray(var_matrix):
#                             var_matrix = list(var_matrix)
#                         data.append(var_matrix)

#                     data = np.array(data)

#                 elif operation['function'] == 'calculate_wind': # to do
#                     pressure   = [v for v in variables if v['key'] == 'pressure'][0]
#                     pressure_name = pressure['name']

#                     u_name = [v['name'] for v in variables if v['key'] == 'u'][0]
#                     v_name = [v['name'] for v in variables if v['key'] == 'v'][0]

#                     time_idxs = list(range(5))
#                     level = [p['value'] for p in operation['parameters'] if operation['parameters']['name'] == 'level_0'][0]

#                     for tidx in time_idxs:
#                         pressure = getvar(ncData, pressure_name, timeidx=tidx,
#                                     squeeze=False, meta=False)
                        
#                         u = getvar(ncData, u_name, timeidx=tidx,
#                                     squeeze=False, meta=False)
                        
#                         v = getvar(ncData, v_name, timeidx=tidx,
#                                     squeeze=False, meta=False)
                        
#                         # to do

#                 elif operation['function'] == 'calculate_hdiv':
#                     pressure   = [v for v in variables if v['key'] == 'pressure'][0]
#                     pressure_name = pressure['name']

#                     u_name = [v['name'] for v in variables if v['key'] == 'u'][0]
#                     v_name = [v['name'] for v in variables if v['key'] == 'v'][0]

#                     time_idxs = list(range(5))
#                     level = [p['value'] for p in operation['parameters'] if operation['parameters']['name'] == 'level_0'][0]
#                     r = [p['value'] for p in operation['parameters'] if operation['parameters']['name'] == 'r'][0]
                    
#                     for t in range(len(time_idxs)):
#                         u = getvar(ncData, u_name, timeidx=t, squeeze=False, meta=False)
#                         v = getvar(ncData, v_name, timeidx=t, squeeze=False, meta=False)
#                         pressure = getvar(ncData, pressure_name, timeidx=t,
#                                 squeeze=False, meta=False)
                        
#                         u_interp = to_np(interplevel(u, pressure, level))
#                         v_interp = to_np(interplevel(v, pressure, level))

#                         hdiv = hdivg(u_interp, v_interp, lat_arr, lon_arr, r)
#                         data.append(hdiv)
                             
#                 elif operation['function'] == 'calculate_kidx':
#                     pressure   = [v for v in variables if v['key'] == 'pressure'][0]
#                     pressure_name = pressure['name']

#                     tc_name = [v for v in variables if v['key'] == 'tc'][0]
#                     rh_name = [v for v in variables if v['key'] == 'rh'][0]

#                     level_0 = [p['value'] for p in operation['parameters'] if operation['parameters']['name'] == 'level_0' ]
#                     level_1 = [p['value'] for p in operation['parameters'] if operation['parameters']['name'] == 'level_1' ]
#                     level_2 = [p['value'] for p in operation['parameters'] if operation['parameters']['name'] == 'level_2' ]

#                     a = [p['value'] for p in operation['parameters'] if operation['parameters']['name'] == 'a']
#                     b = [p['value'] for p in operation['parameters'] if operation['parameters']['name'] == 'b']
#                     mv = [p['value'] for p in operation['parameters'] if operation['parameters']['name'] == 'mv']
#                     le = [p['value'] for p in operation['parameters'] if operation['parameters']['name'] == 'le']
#                     r = [p['value'] for p in operation['parameters'] if operation['parameters']['name'] == 'r']

#                     for t in range(len(time_idxs)):
                    
#                         tc = getvar(ncData, tc_name, timeidx=t, squeeze=False, meta=False)
#                         rh = getvar(ncData, rh_name, timeidx=t, squeeze=False, meta=False)
#                         pressure = getvar(ncData, pressure_name, timeidx=t, squeeze=False, meta=False)

#                         tc_interp_0 = to_np(interplevel(tc, pressure, level_0))
#                         tc_interp_1 = to_np(interplevel(tc, pressure, level_1))
#                         tc_interp_2 = to_np(interplevel(tc, pressure, level_2))

#                         rh_interp_0 =  to_np(interplevel(tc, pressure, level_0))
#                         rh_interp_1 =  to_np(interplevel(tc, pressure, level_1))

#                         arr = []

#                         for i in range(len(lat_arr)):
#                             row_arr = []

#                             for j in range(len(lon_arr)):
#                                 t1 = tc_interp_0[i][j]
#                                 t2 = tc_interp_1[i][j]
#                                 t4 = tc_interp_2[i][j]

#                                 rh_0 = rh_interp_0[i][j]
#                                 rh_1 = rh_interp_1[i][j]

#                                 e3 = a * math.exp((mv*le/r) * ((1/b) - (1/(t1+b))))
#                                 e5 = a * math.exp((mv*le/r) * ((1/b) - (1/(t4+b))))
                                
#                                 es3 = rh_0[i][j]*e3/100
#                                 es5 = rh_1[i][j]*e5/100

#                                 td3 = 1/(1/b-(r/(mv*le)) * math.log10(es3/a)) - k
#                                 td5 = 1/(1/b-(r/(mv*le)) * math.log10(es5/a)) - k

#                                 t3 = td3
#                                 t5 = td5

#                                 k = t1-t2+t3-t4+t5

#                                 row_arr.append(k)
#                             arr.append(row_arr)
#                         data.append(arr)
#                     data = np.array(data)
            
#             elif operation['dimension'] == 'time':
#                 if operation['function'] == 'accumulate':
#                     interval = operation['parameters'][0]
#                     data = [data[t] + data[t + interval] for t in range(0, len(data), interval)]
                
#                 elif operation['function'] == 'unaccumulate':
#                     interval = operation['parameters'][0]
#                     data = [data[t] - data[t - interval] for t in range(len(data), 0-1, interval)]

#                 elif operation['function'] == 'avg':
#                     data = [data[t] - data[t - interval] for t in range(len(data), 0-1, interval)]

#         elif operation['type'] == 'vector2D_to_vector1D':
#             print('to do')
            
#         elif operation['type'] == 'scalar':
#             if len(data) == 0:
#                 data = ncData.variables[variables[0]['name']][:]

#             for matrix in data:
#                 for row in matrix:
#                     for i in range(len(row)):
#                         apply_scalar = eval(operation['function'])
#                         row[i] = float(round(apply_scalar(row[i], operation['value']), 2))

#     ############################################################################################

#     points = []
#     values = []

#     for latidx in range(min_lat_idx, max_lat_idx + 1):
#         for lonidx in range(min_lon_idx, max_lon_idx + 1):
#             points.append((lat_arr[latidx], lon_arr[lonidx]))
#             pt_arr = []
            
#             for tidx in range(len(data)):
#                 pt_arr.append(float(data[tidx][latidx][lonidx]))

#             values.append(pt_arr)

#     coordinates = []
#     transformer = Transformer.from_crs(coords['proj'], 3395)

#     for point in transformer.itransform(points):
#         coordinates.append(float(point[0]))
#         coordinates.append(float(point[1]))
#         coordinates.append(0)

#     abstract_json = {
#         "id": layer_id,
#         "coordinates": coordinates,
#         "values": values
#     }

#     json_object = json.dumps(abstract_json)

#     directory = os.path.dirname(file_path)

#     with open(os.path.join(directory,layer_id+".json"), "w") as outfile:
#         outfile.write(json_object)

'''
    Thematic data from numpy array file 

    coordinates shape: (n,3)
    Considers that coordinates do not have a coordinates system but are in meters
'''
def thematic_from_npy(filepath_coordinates, filepath_values, layer_id, center_around=[]):

    coordinates = np.load(filepath_coordinates)
    values = np.load(filepath_values)

    coordinates = coordinates.flatten()

    if(len(center_around) > 0):
        coordinates = center_coordinates_around(coordinates, center_around)

    flat_values = []

    if(isinstance(values[0], np.ndarray)):
        flat_values = [item for row in values for item in row] 
    else:
        flat_values = values.tolist()

    abstract_json = {
        "id": layer_id,
        "coordinates": coordinates.tolist(),
        "values": flat_values
    }

    json_object = json.dumps(abstract_json)
    
    directory = os.path.dirname(filepath_coordinates)

    with open(os.path.join(directory,layer_id+".json"), "w") as outfile:
        outfile.write(json_object)


import pandas as pd
import geopandas as gpd

gdf_points = arg # Getting GeoDataFrame from previous node

gdf_points = gdf_points.to_crs(3857) # Transforming to a a coordinate system that operates in meters

gdf_points["geometry"] = gdf_points["geometry"].buffer(25)

gdf_points = gdf_points.to_crs(4326) # Transforming back to Latitude and Longitude

return gdf_points


import geopandas as gpd
import pandas as pd
from shapely.geometry import Point, Polygon, box

from . import lineclipping

def convert_projections(inProj, outProj, geometry, dim2=True):
    '''
        Convert projections from inProj to outPorj. Both values can be anything accepted by pyproj.CRS.from_user_input(), such as an authority string (eg “EPSG:4326”) or a WKT string.
        The functions converts the geometries into a geodataframe to make the conversion faster using geopandas.

        *The input follows the x,y format. For instance, if the input projection 4326 a point [longitude, latitude] is expected*

        * @param {string} inProj The current projection of the geometry
        * @param {string} outProj The desired out projection for the geometry
        * @param {bool} Indicates if the geometry is represented by 2D points or 3D points
        * @returns {List} Returns a list with all geometries transformed
    '''
    # transform the flat array into a list of Points
    grouped_geometry = []

    index = 0
    offset = 1 if dim2 else 2
    while(index < len(geometry)-offset):
        if dim2:
            grouped_geometry.append(Point(geometry[index+1], geometry[index])) # must be long/lat
        else:
            grouped_geometry.append(Point(geometry[index+1], geometry[index], geometry[index+2])) # must be long/lat

        index += 2 if dim2 else 3

    df = pd.DataFrame(grouped_geometry)

    gdf = gpd.GeoDataFrame(df, geometry=0, crs=inProj)

    gdf = gdf.to_crs(outProj)   

    translatedGeometry = []

    for elem in gdf[0]:
        translatedGeometry.append(elem.x)
        translatedGeometry.append(elem.y)
        if not dim2:
            translatedGeometry.append(elem.z)

    return translatedGeometry

def get_camera(coordinates, bbox=False):

    center = [(0,0)]

    if(bbox):
        center = [(coordinates[0]+coordinates[2])/2.0,(coordinates[1]+coordinates[3])/2.0]
    else:
        polygon = polygon_bpoly(coordinates, bbox)
        center = list(polygon.centroid.coords[0])

    center = convert_projections("4326", "3395", center)
    center.append(1) # zoom level

    return {
        'position': center, 
        'direction': {
            'right': [0,0,3000],
            'lookAt': [0,0,0],
            'up': [0,1,0]
        }
    }

def intersect_bbox(bb1, bb2):
    if point_within_bbox([bb1[0],bb1[1]],bb2) or point_within_bbox([bb1[2],bb1[3]],bb2)\
        or point_within_bbox([bb1[0],bb1[3]],bb2) or point_within_bbox([bb1[2],bb1[1]],bb2):
        return True
    return False

def point_within_bbox(point, bb):
    if point[0] > bb[0] and point[0] < bb[2] and point[1] > bb[1] and point[1] < bb[3]:
        return True
    return False

def intersect_line_bbox(p0, p1, bb):
    result = lineclipping.cohensutherland(bb[1],bb[2],bb[3],bb[0],p0[1],p0[0],p1[1],p1[0])
    return result


def extend_bbox(bounding_box,latlng):
    bbox = bounding_box.copy()
    bbox[0]=min(bbox[0]+90.0,latlng[0]+90.0)-90.0
    bbox[1]=min(bbox[1]+180.0,latlng[1]+180.0)-180.0
    bbox[2]=max(bbox[2]+90.0,latlng[0]+90.0)-90.0
    bbox[3]=max(bbox[3]+180.0,latlng[1]+180.0)-180.0
    return bbox

# triangulation deviation
# code from: https://github.com/joshuaskelly/earcut-python
# Copyright (c) 2016, Mapbox
# ISC License
def signedArea(data, start, end, dim):
    sum = 0
    j = end - dim

    for i in range(start, end, dim):
        sum += (data[j] - data[i]) * (data[i + 1] + data[j + 1])
        j = i

    return sum
    
def deviation(data, holeIndices, dim, triangles):
    _len = len(holeIndices)
    hasHoles = holeIndices and len(holeIndices)
    outerLen = holeIndices[0] * dim if hasHoles else len(data)

    polygonArea = abs(signedArea(data, 0, outerLen, dim))

    if hasHoles:
        for i in range(_len):
            start = holeIndices[i] * dim
            end = holeIndices[i + 1] * dim if i < _len - 1 else len(data)
            polygonArea -= abs(signedArea(data, start, end, dim))

    trianglesArea = 0

    for i in range(0, len(triangles), 3):
        a = triangles[i] * dim
        b = triangles[i + 1] * dim
        c = triangles[i + 2] * dim
        trianglesArea += abs(
            (data[a] - data[c]) * (data[b + 1] - data[a + 1]) -
            (data[a] - data[b]) * (data[c + 1] - data[a + 1]))

    if polygonArea == 0 and trianglesArea == 0:
        return 0

    return abs((trianglesArea - polygonArea) / polygonArea)

def from_2d_to_3d(nodes, z_offset=0):
    '''
        Inserts a z position (z_offset can be applied)
    '''
    new_3d_node = []

    index = 0
    while(index < len(nodes)-1):
        new_3d_node.append(nodes[index])
        new_3d_node.append(nodes[index+1])
        new_3d_node.append(0 + z_offset)
        index += 2

    return new_3d_node

def polygon_bpoly(coordinates, bbox = False):

    if(not bbox):
        groupped_coords = []

        for i in range(int(len(coordinates)/2)):
            groupped_coords.append((coordinates[i*2], coordinates[i*2+1]))

        groupped_coords.append((coordinates[-2], coordinates[-1])) # close the polygon

        return Polygon(groupped_coords)
    else:
        return box(coordinates[0], coordinates[1], coordinates[2], coordinates[3])

'''
    coordinates is flattened x,y,z
    center_around is [x,y,z]
'''
def center_coordinates_around(coordinates, center_around, scale_up = 1):
    xmin = None
    xmax = None
    ymin = None
    ymax = None
    zmin = None
    zmax = None

    for i in range(int(len(coordinates)/3)):
        
        if(xmin == None or coordinates[i*3] < xmin):
            xmin = coordinates[i*3]

        if(xmax == None or coordinates[i*3] > xmax):
            xmax = coordinates[i*3]

        if(ymin == None or coordinates[i*3+1] < ymin):
            ymin = coordinates[i*3+1]

        if(ymax == None or coordinates[i*3+1] > ymax):
            ymax = coordinates[i*3+1]

        if(zmin == None or coordinates[i*3+2] < zmin):
            zmin = coordinates[i*3+2]

        if(zmax == None or coordinates[i*3+2] > zmax):
            zmax = coordinates[i*3+2]

    current_center = [(xmin+xmax)/2, (ymin+ymax)/2, (zmin+zmax)/2]

    # centering data around 0
    for i in range(int(len(coordinates)/3)):
        coordinates[i*3] -= current_center[0]
        coordinates[i*3] *= scale_up # scalling objects up
        coordinates[i*3+1] -= current_center[1]
        coordinates[i*3+1] *= scale_up # scalling objects up
        coordinates[i*3+2] -= current_center[2]
        coordinates[i*3+2] *= scale_up # scalling objects up

    # center data around new center
    for i in range(int(len(coordinates)/3)):
        coordinates[i*3] += center_around[0]
        coordinates[i*3+1] += center_around[1]
        coordinates[i*3+2] += center_around[2]
        
    return coordinates
